###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for LHCbDIRAC utilities."""
# pylint: disable=protected-access,missing-docstring,invalid-name
import unittest
import itertools
import os
import datetime

from unittest.mock import MagicMock

from DIRAC import gLogger

from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

from LHCbDIRAC.Core.Utilities.ProductionData import (
    _makeProductionLFN,
    constructProductionLFNs,
    _getLFNRoot,
    _applyMask,
    getLogPath,
    constructUserLFNs,
)
from LHCbDIRAC.Core.Utilities.InputDataResolution import InputDataResolution
from LHCbDIRAC.Core.Utilities.GangaDataFile import GangaDataFile
from LHCbDIRAC.Core.Utilities.NagiosConnector import NagiosConnector

gConfigMock = MagicMock()
gConfigMock.getValue.return_value = "aValue"


class UtilitiesTestCase(unittest.TestCase):
    """Base class for the Utilities test cases."""

    def setUp(self):
        self.IDR = InputDataResolution({}, bkc_mock)
        gLogger.setLevel("DEBUG")

    def tearDown(self):
        for fileProd in ["data.py", "gaudi_extra_options.py"]:
            try:
                os.remove(fileProd)
            except OSError:
                continue


#################################################


class GangaDataFileSuccess(UtilitiesTestCase):
    def test_generateGangaDataFile(self):
        gdf = GangaDataFile()

        res = gdf.generateDataFile(["foo", "bar"], persistency="ROOT")
        # Remove first line as it contains the date and time
        res = "\n".join(res.split("\n")[1:])
        root = (
            "\n"
            + "from Gaudi.Configuration import * \n"
            + "from GaudiConf import IOHelper\nIOHelper('ROOT').inputFiles([\n"
            + "'LFN:foo',\n'LFN:bar',\n], clear=True)\n"
            + "\nFileCatalog().Catalogs += [ 'xmlcatalog_file:pool_xml_catalog.xml' ]\n"
        )
        self.assertEqual(res, root)

        res = gdf.generateDataFile(["foo", "bar"])
        # Remove first line as it contains the date and time
        res = "\n".join(res.split("\n")[1:])
        root = (
            "\n"
            + "from Gaudi.Configuration import * \n"
            + "from GaudiConf import IOHelper\nIOHelper().inputFiles([\n"
            + "'LFN:foo',\n'LFN:bar',\n], clear=True)\n"
            + "\nFileCatalog().Catalogs += [ 'xmlcatalog_file:pool_xml_catalog.xml' ]\n"
        )
        self.assertEqual(res, root)

        gdf = GangaDataFile(xmlcatalog_file="")

        res = gdf.generateDataFile(["foo", "bar"], persistency="MDF")
        # Remove first line as it contains the date and time
        res = "\n".join(res.split("\n")[1:])
        root = (
            "\n"
            + "from Gaudi.Configuration import * \n"
            + "from GaudiConf import IOHelper\nIOHelper('MDF').inputFiles([\n"
            + "'LFN:foo',\n'LFN:bar',\n], clear=True)\n"
        )
        self.assertEqual(res, root)

        res = gdf.generateDataFile("foo", persistency="ROOT")
        # Remove first line as it contains the date and time
        res = "\n".join(res.split("\n")[1:])
        root = (
            "\n"
            + "from Gaudi.Configuration import * \n"
            + "from GaudiConf import IOHelper\nIOHelper('ROOT').inputFiles([\n"
            + "'LFN:foo',\n], clear=True)\n"
        )
        self.assertEqual(res, root)


#################################################


class GangaDataFileFailure(UtilitiesTestCase):
    def test_generateDataFile(self):
        gdf = GangaDataFile()
        self.assertRaises(TypeError, gdf.generateDataFile, (""))
        self.assertRaises(ValueError, gdf.generateDataFile, ([]))


#################################################


class ProductionDataSuccess(UtilitiesTestCase):
    #################################################

    def test_constructProductionLFNs(self):
        # + test with InputData

        paramDict = {
            "PRODUCTION_ID": "12345",
            "JOB_ID": "54321",
            "configVersion": "test",
            "configName": "certification",
            "JobType": "MCSimulation",
            "outputList": [
                {"outputDataType": "sim", "outputDataSE": "Tier1-RDST", "outputDataName": "00012345_00054321_1.sim"},
                {"outputDataType": "digi", "outputDataSE": "Tier1-RDST", "outputDataName": "00012345_00054321_2.digi"},
                {
                    "outputDataType": "dst",
                    "outputDataSE": "Tier1_MC_M-DST",
                    "outputDataName": "00012345_00054321_4.dst",
                },
                {
                    "outputDataType": "ALLSTREAMS.DST",
                    "outputBKType": "ALLSTREAMS.DST",
                    "outputDataSE": "Tier1_MC_M-DST",
                    "outputDataName": "00012345_00054321_5.AllStreams.dst",
                },
            ],
        }

        AllStreamsDST = "/lhcb/certification/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst"
        reslist = [
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
            },
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": ["/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst"],
            },
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": ["/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst"],
            },
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": [
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                ],
            },
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": [AllStreamsDST],
            },
            {
                "LogTargetPath": ["/lhcb/certification/test/LOG/00012345/0005/00054321.zip"],
                "LogFilePath": ["/lhcb/certification/test/LOG/00012345/0005"],
                "DebugLFNs": [
                    "/lhcb/debug/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/debug/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/debug/test/DST/00012345/0005/00012345_00054321_4.dst",
                    "/lhcb/debug/test/ALLSTREAMS.DST/00012345/0005/00012345_00054321_5.AllStreams.dst",
                    "/lhcb/debug/test/CORE/00012345/0005/00054321_core",
                ],
                "BookkeepingLFNs": [
                    "/lhcb/certification/test/SIM/00012345/0005/00012345_00054321_1.sim",
                    "/lhcb/certification/test/DIGI/00012345/0005/00012345_00054321_2.digi",
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
                "ProductionOutputData": [
                    "/lhcb/certification/test/DST/00012345/0005/00012345_00054321_4.dst",
                    AllStreamsDST,
                ],
            },
        ]

        outputDataFileMasks = ("", "dst", "DST", ["digi", "dst"], "ALLSTREAMS.DST", ["dst", "allstreams.dst"])

        for outputDataFileMask, resL in zip(outputDataFileMasks, reslist):
            paramDict["outputDataFileMask"] = outputDataFileMask

            res = constructProductionLFNs(paramDict, bkc_mock)

            self.assertTrue(res["OK"])
            self.assertEqual(res["Value"], resL)

            resWithBkk = constructProductionLFNs(paramDict, bkc_mock, quick=False)

            self.assertTrue(resWithBkk["OK"])
            self.assertEqual(resWithBkk["Value"], resL)

    #################################################

    def test__makeProductionLFN(self):
        JOB_ID = "00054321"
        LFN_ROOT = "/lhcb/MC/MC10"
        filetuple = (("00012345_00054321_1.sim", "sim"), ("00012345_00054321_1.sim", "SIM"))
        prodstring = "00012345"

        for ft in filetuple:
            res = _makeProductionLFN(JOB_ID, LFN_ROOT, ft, prodstring)
            self.assertEqual(res, "/lhcb/MC/MC10/SIM/00012345/0005/00012345_00054321_1.sim")

    #################################################

    def test__applymask(self):
        dtl = [
            ("00012345_00054321_1.sim", "sim"),
            ("00012345_00054321_4.dst", "dst"),
            ("00012345_00054321_2.digi", "digi"),
            ("Brunel_00012345_00012345_1_Hist.root", "hist"),
            ("00012345_00054321_5.AllStreams.dst", "ALLSTREAMS.DST"),
        ]

        wfMask = ("", "dst", "ALLSTREAMS.DST", ["dst", "digi"], ["DIGI", "allstreams.dst"], "hist", ["dst", "hist"])

        dtlM = (
            [
                ("00012345_00054321_1.sim", "sim"),
                ("00012345_00054321_4.dst", "dst"),
                ("00012345_00054321_2.digi", "digi"),
                ("Brunel_00012345_00012345_1_Hist.root", "hist"),
                ("00012345_00054321_5.AllStreams.dst", "ALLSTREAMS.DST"),
            ],
            [("00012345_00054321_4.dst", "dst")],
            [("00012345_00054321_5.AllStreams.dst", "ALLSTREAMS.DST")],
            [("00012345_00054321_4.dst", "dst"), ("00012345_00054321_2.digi", "digi")],
            [("00012345_00054321_2.digi", "digi"), ("00012345_00054321_5.AllStreams.dst", "ALLSTREAMS.DST")],
            [("Brunel_00012345_00012345_1_Hist.root", "hist")],
            [("00012345_00054321_4.dst", "dst"), ("Brunel_00012345_00012345_1_Hist.root", "hist")],
        )

        for mask, res in zip(wfMask, dtlM):
            r = _applyMask(mask, dtl)

            self.assertEqual(r, res)

    def test_getLogPath(self):
        wkf_commons = {
            "PRODUCTION_ID": 12345,
            "JOB_ID": 0o0001,
            "configName": "LHCb",
            "configVersion": "Collision11",
            "JobType": "MCSimulation",
        }

        resWithBkk = getLogPath(wkf_commons, bkc_mock, quick=False)

        self.assertEqual(
            resWithBkk,
            {
                "OK": True,
                "Value": {
                    "LogTargetPath": ["/lhcb/LHCb/Collision11/LOG/00012345/0000/00000001.zip"],
                    "LogFilePath": ["/lhcb/LHCb/Collision11/LOG/00012345/0000"],
                },
            },
        )

        res = getLogPath(wkf_commons, bkc_mock)

        self.assertEqual(
            res,
            {
                "OK": True,
                "Value": {
                    "LogTargetPath": ["/lhcb/LHCb/Collision11/LOG/00012345/0000/00000001.zip"],
                    "LogFilePath": ["/lhcb/LHCb/Collision11/LOG/00012345/0000"],
                },
            },
        )

    def test__getLFNRoot(self):
        res = _getLFNRoot("/lhcb/data/CCRC08/00009909/DST/0000/00009909_00003456_2.dst", "MC12", bkClient=bkc_mock)
        self.assertEqual(res, "/lhcb/data/CCRC08/00009909/DST/0000/00009909_00003456_2.dst")

        res = _getLFNRoot(
            "/lhcb/data/CCRC08/00009909/DST/0000/00009909_00003456_2.dst", "MC12", bkClient=bkc_mock, quick=True
        )
        self.assertEqual(res, "/lhcb/data/CCRC08/00009909")

    def test_constructUserLFNs(self):
        timeTup = datetime.date.today().timetuple()
        yearMonth = f"{timeTup[0]}_{str(timeTup[1]).zfill(2)}"

        res = constructUserLFNs(123, "fstagni", ["pippo.txt", "pluto.txt"])
        self.assertEqual(
            sorted(res),
            sorted(
                [
                    "/lhcb/user/f/fstagni/" + yearMonth + "/0/123/pluto.txt",
                    "/lhcb/user/f/fstagni/" + yearMonth + "/0/123/pippo.txt",
                ]
            ),
        )

        res = constructUserLFNs(123, "fstagni", ["pippo.txt", "pluto.txt"], "my/Output/Path")
        self.assertEqual(
            sorted(res),
            sorted(
                [
                    "/lhcb/user/f/fstagni/my/Output/Path/" + yearMonth + "/0/123/pluto.txt",
                    "/lhcb/user/f/fstagni/my/Output/Path/" + yearMonth + "/0/123/pippo.txt",
                ]
            ),
        )

        res = constructUserLFNs(123, "fstagni", ["pippo.txt", "pluto.txt"], prependString="pre/pend")
        self.assertEqual(
            sorted(res),
            sorted(
                [
                    "/lhcb/user/f/fstagni/" + yearMonth + "/123_pre/pend_pippo.txt",
                    "/lhcb/user/f/fstagni/" + yearMonth + "/123_pre/pend_pluto.txt",
                ]
            ),
        )

        res = constructUserLFNs(123, "fstagni", ["pippo.txt", "pluto.txt"], "my/Output/Path", "pre/pend")
        self.assertEqual(
            sorted(res),
            sorted(
                [
                    "/lhcb/user/f/fstagni/my/Output/Path/" + yearMonth + "/123_pre/pend_pippo.txt",
                    "/lhcb/user/f/fstagni/my/Output/Path/" + yearMonth + "/123_pre/pend_pluto.txt",
                ]
            ),
        )


class InputDataResolutionSuccess(UtilitiesTestCase):
    def test__addPfnType(self):
        res = self.IDR._addPfnType({"lfn1": {"mdata": "mdata1"}, "lfn2": {"mdata": "mdata2"}})
        self.assertEqual(
            res,
            {
                "OK": True,
                "Value": {
                    "lfn1": {"pfntype": "ROOT", "mdata": "mdata1"},
                    "lfn2": {"pfntype": "MDF", "mdata": "mdata2"},
                },
            },
        )

        bkc_mock.getFileTypeVersion.return_value = {"OK": True, "Value": {}}

        res = self.IDR._addPfnType({"lfn1": {"mdata": "mdata1"}, "lfn2": {"mdata": "mdata2"}})
        self.assertEqual(
            res,
            {
                "OK": True,
                "Value": {
                    "lfn1": {"pfntype": "ROOT", "mdata": "mdata1"},
                    "lfn2": {"pfntype": "ROOT", "mdata": "mdata2"},
                },
            },
        )

        bkc_mock.getFileTypeVersion.return_value = {"OK": True, "Value": {"lfn1": "ROOT"}}

        res = self.IDR._addPfnType(
            {"lfn1": {"mdata": "mdata1"}, "lfn2": {"mdata": "mdata2"}, "lfn3": {"mdata": "mdata3"}}
        )
        self.assertEqual(
            res,
            {
                "OK": True,
                "Value": {
                    "lfn1": {"pfntype": "ROOT", "mdata": "mdata1"},
                    "lfn2": {"pfntype": "ROOT", "mdata": "mdata2"},
                    "lfn3": {"pfntype": "ROOT", "mdata": "mdata3"},
                },
            },
        )


class NagiosConnectorSuccess(UtilitiesTestCase):
    def test_readConfig(self):
        nagConn = NagiosConnector()
        nagConn.readConfig()
        self.assertTrue(nagConn.config["MsgPort"])


#  def test_failedConnection( self ):
#
#    nagConn = NagiosConnector()
#    nagConn.readConfig()
#    nagConn.config['NagiosName'] = 'lhcb-Dirac.Unittest'
#    nagConn.useDebugMessage()
#    self.assertEqual( nagConn.initializeConnection(),
#                       S_OK( 'Connection to Broker established' ),
#                       'Connection not correctly initialized' )
#    self.assertEqual( nagConn.sendMessage(),
#                       S_OK('Message sent to Broker.'),
#                       'Sending unsuccessful!' )


#############################################################################
# Test Suite run
#############################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(UtilitiesTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ProductionDataSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(InputDataResolutionSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(NagiosConnectorSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Moving toward a templates-less system.

The RecoStripping Template creates workflows for the following use-cases:
  WORKFLOW1: Reconstruction
  WORKFLOW2: Stripping+Merge
  WORKFLOW3: RecoStripping+Merge (Reco and Stripping within the same job)
  WORKFLOW4: Reconstruction+Stripping+Merge
  WORKFLOW5: Stripping+Merge+Indexing

Exotic things you might want to do:
* run a local test:
  pre: remember to check if your input file is online, if not use lcg-bringonline <PFN>
* run only part of the request on the Grid:
"""

import ast

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

gLogger = gLogger.getSubLogger("RecoStripping_run.py")
currentSetup = gConfig.getValue("DIRAC/Setup")

pr = ProductionRequest()

stepsList = ["{{p1Step}}"]
stepsList.append("{{p2Step}}")
stepsList.append("{{p3Step}}")
stepsList.append("{{p4Step}}")
stepsList.append("{{p5Step}}")
stepsList.append("{{p6Step}}")
stepsList.append("{{p7Step}}")
stepsList.append("{{p8Step}}")
stepsList.append("{{p9Step}}")
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = "{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}"

w = "{{w#----->WORKFLOW: choose one below#}}"
w1 = "{{w1#-WORKFLOW1: Reconstruction#False}}"
w2 = "{{w2#-WORKFLOW2: Stripping+Merge#False}}"
w3 = "{{w3#-WORKFLOW3: RecoStripping+Merge#False}}"
w4 = "{{w4#-WORKFLOW4: Reconstruction+Stripping+Merge#False}}"
w5 = "{{w5#-WORKFLOW5: Stripping+Merge+Indexing #False}}"

localTestFlag = "{{localTestFlag#GENERAL: Set True for local test#False}}"
validationFlag = "{{validationFlag#GENERAL: Set True for validation prod#False}}"

# workflow params for all productions
pr.startRun = "{{startRun#GENERAL: run start, to set the start run#0}}"
pr.endRun = "{{endRun#GENERAL: run end, to set the end of the range#0}}"
pr.runsList = "{{runsList#GENERAL: discrete list of run numbers (do not mix with start/endrun)#}}"
pr.derivedProduction = "{{AncestorProd#GENERAL: ancestor prod to be derived#0}}"

# reco params
recoPriority = int("{{RecoPriority#PROD-1:RECO(Stripp): priority#2}}")
recoCPU = "{{RecoMaxCPUTime#PROD-1:RECO(Stripp): Max CPU time in secs#1500000}}"
recoPlugin = "{{RecoPluginType#PROD-1:RECO(Stripp): production plugin name#RAWProcessing}}"
recoFilesPerJob = "{{RecoFilesPerJob#PROD-1:RECO(Stripp): Group size or number of files per job#1}}"
recoDataSE = "{{RecoDataSE#PROD-1:RECO(Stripp): Output Data Storage Element#Tier1-Buffer}}"
try:
    recoDataSESpecial = ast.literal_eval(
        "{{RecoDataSESpecial#PROD-1:RECO(Stripp): Special SE (a dictionary Type:SE)#}}"
    )
except SyntaxError:
    recoDataSESpecial = {}
recoAncestorDepth = int("{{recoAncestorDepth#PROD-1: Ancestor Depth#0}}")
recoCompressionLvl = "{{recoCompressionLvl#PROD-1: compression level#LOW}}"
recoOutputVisFlag = "{{recoOutputVisFlag#PROD-1: Visibility flag of output files #Y}}"
try:
    recoOutputVisFlagSpecial = ast.literal_eval(
        "{{recoOutputVisFlagSpecial#PROD-1: Special Visibility flag of output files (dict FType:Y|N )#}}"
    )
except SyntaxError:
    recoOutputVisFlagSpecial = {}

# stripp params
strippPriority = int("{{priority#PROD-2:Stripping: priority#2}}")
strippCPU = "{{StrippMaxCPUTime#PROD-2:Stripping: Max CPU time in secs#1000000}}"
strippPlugin = "{{StrippPluginType#PROD-2:Stripping: plugin name#ByRunWithFlush}}"
strippFilesPerJob = "{{StrippFilesPerJob#PROD-2:Stripping: Group size or number of files per job#2}}"
strippDataSE = "{{StrippStreamSE#PROD-2:Stripping: output data SE (un-merged streams)#Tier1-Buffer}}"
try:
    strippDataSESpecial = ast.literal_eval(
        "{{StrippDataSESpecial#PROD-2:Stripping: Special SE (a dictionary Type:SE)#}}"
    )
except SyntaxError:
    strippDataSESpecial = {}
strippAncestorDepth = int("{{strippAncestorDepth#PROD-2: Ancestor Depth#0}}")
strippCompressionLvl = "{{strippCompressionLvl#PROD-2: compression level#LOW}}"
strippOutputVisFlag = "{{strippOutputVisFlag#PROD-2: Visibility flag of output files#N}}"
try:
    strippOutputVisFlagSpecial = ast.literal_eval(
        "{{strippOutputVisFlagSpecial#PROD-2: Special Visibility flag of output files (dict FType:Y|N)#}}"
    )
except SyntaxError:
    strippOutputVisFlagSpecial = {}

# merging params
mergingPriority = int("{{MergePriority#PROD-3:Merging: priority#8}}")
mergingCPU = "{{MergeMaxCPUTime#PROD-3:Merging: Max CPU time in secs#300000}}"
mergingPlugin = "{{MergePlugin#PROD-3:Merging: plugin#ByRunFileTypeSizeWithFlush}}"
mergingGroupSize = "{{MergeFileSize#PROD-3:Merging: Size (in GB) of the merged files#5}}"
mergingDataSE = "{{MergeStreamSE#PROD-3:Merging: output data SE (merged streams)#Tier1-DST}}"
try:
    mergingDataSESpecial = ast.literal_eval(
        "{{MergingDataSESpecial#PROD-3:Merging: Special SE (a dictionary Type:SE)#}}"
    )
except SyntaxError:
    mergingDataSESpecial = {}
mergingRemoveInputsFlag = "{{MergeRemoveFlag#PROD-3:Merging: remove input data flag True/False#True}}"
mergeCompressionLvl = "{{mergeCompressionLvl#PROD-3: compression level#HIGH}}"
mergeOutputVisFlag = "{{mergeOutputVisFlag#PROD-3: Visibility flag of output files#Y}}"
try:
    mergeOutputVisFlagSpecial = ast.literal_eval(
        "{{mergeOutputVisFlagSpecial#PROD-3: Special Visibility flag of output files (dict FType:Y|N)#}}"
    )
except SyntaxError:
    mergeOutputVisFlagSpecial = {}

# indexing params
indexingPriority = int("{{IndexingPriority#PROD-4:indexing: priority#5}}")
indexingCPU = "{{IndexingMaxCPUTime#PROD-4:indexing: Max CPU time in secs#50000}}"
indexingPlugin = "{{IndexingPlugin#PROD-4:indexing: plugin#ByRunFileTypeSizeWithFlush}}"
indexingGroupSize = "{{IndexingFileSize#PROD-4:indexing: Size (in GB) of the Indexed files#50}}"
indexingDataSE = "{{IndexingStreamSE#PROD-4:indexing: output data SE (Indexed streams)#IndexerSE}}"

pr.requestID = "{{ID}}"
pr.prodGroup = "{{inProPass}}" + "/" + "{{pDsc}}"
# used in case of a test e.g. certification etc.
pr.configName = "{{configName}}"
pr.configVersion = "{{configVersion}}"
# Other parameters from the request page
pr.dqFlag = "{{inDataQualityFlag}}"  # UNCHECKED
pr.dataTakingConditions = "{{simDesc}}"
pr.processingPass = "{{inProPass}}"
pr.bkFileType = "{{inFileType}}"
pr.eventType = "{{eventType}}"
pr.visibility = "Yes"
recoType = "DataReconstruction"
targetSite = "ALL"
recoMulticoreFlag = strippMulticoreFlag = mergeMulticoreFlag = "True"
recoIDPolicy = strippIDPolicy = mergingIDPolicy = indexingIDPolicy = "download"


w1 = ast.literal_eval(w1)
w2 = ast.literal_eval(w2)
w3 = ast.literal_eval(w3)
w4 = ast.literal_eval(w4)
w5 = ast.literal_eval(w5)

localTestFlag = ast.literal_eval(localTestFlag)
validationFlag = ast.literal_eval(validationFlag)

mergeRemoveInputsFlag = ast.literal_eval(mergingRemoveInputsFlag)

if not w1 and not w2 and not w3 and not w4:
    gLogger.error("I told you to select at least one workflow!")
    DIRACexit(2)

if localTestFlag:
    pr.testFlag = True
    pr.publishFlag = False
    pr.prodsToLaunch = [1]

recoInputDataList = []
strippInputDataList = []
if not pr.publishFlag:
    # this is 1380Gev MagUp
    # recoTestData = 'LFN:/lhcb/data/2011/RAW/FULL/LHCb/COLLISION11/88162/088162_0000000020.raw'
    # this is collision11
    recoTestData = "LFN:/lhcb/data/2011/RAW/FULL/LHCb/COLLISION11/89333/089333_0000000003.raw"
    recoInputDataList.append(recoTestData)
    recoIDPolicy = "protocol"

    strippTestData = "LFN:/lhcb/data/2010/SDST/00008375/0001/00008375_00016947_1.sdst"
    strippInputDataList.append(strippTestData)
    #  strippTestDataRAW = 'LFN:/lhcb/data/2010/RAW/FULL/LHCb/COLLISION10/75338/075338_0000000069.raw'
    #  strippInputDataList.append( strippTestDataRAW )
    strippIDPolicy = "protocol"
    evtsPerJob = "2000"


pr.outConfigName = pr.configName

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == "LHCb-Certification" or pr.testFlag:
    pr.outConfigName = "certification"
    if "lhcb" not in pr.configName.lower():  # in case of reconstruction
        pr.configVersion = "test"

if pr.testFlag:
    pr.dataTakingConditions = "Beam3500GeV-VeloClosed-MagUp"
    if w1 or w3:
        pr.events = ["25"]
        pr.processingPass = "Real Data"
        pr.bkFileType = "RAW"
    else:
        pr.events = ["2000"]
        pr.processingPass = "Real Data/Reco12"
        pr.bkFileType = "SDST"
    mergingGroupSize = "1"
    recoCPU = strippCPU = "200000"
    pr.startRun = "93718"
    pr.endRun = "93720"
    pr.eventType = "90000000"
    pr.dqFlag = "ALL"

if validationFlag:
    pr.outConfigName = "validation"

if w1:
    pr.prodsTypeList = [recoType]
    pr.outputSEs = [recoDataSE]
    pr.specialOutputSEs = [recoDataSESpecial]
    pr.stepsInProds = [list(range(1, len(pr.stepsList) + 1))]
    pr.removeInputsFlags = [False]
    pr.priorities = [recoPriority]
    pr.cpus = [recoCPU]
    pr.groupSizes = [recoFilesPerJob]
    pr.plugins = [recoPlugin]
    pr.inputs = [recoInputDataList]
    pr.inputDataPolicies = [recoIDPolicy]
    pr.bkQueries = ["Full"]
    pr.targets = [targetSite]
    pr.multicore = [recoMulticoreFlag]
    pr.outputModes = ["Run"]
    pr.ancestorDepths = [recoAncestorDepth]
    pr.compressionLvl = [recoCompressionLvl] * len(pr.stepsInProds[0])
    pr.outputVisFlag = [{str(i + 1): recoOutputVisFlag} for i in range(len(pr.stepsInProds[0]))]
    pr.specialOutputVisFlag = [{"1": recoOutputVisFlagSpecial}]

elif w2:
    pr.prodsTypeList = ["DataStripping", "Merge"]
    pr.outputSEs = [strippDataSE, mergingDataSE]
    pr.specialOutputSEs = [strippDataSESpecial, mergingDataSESpecial]
    pr.stepsInProds = [list(range(1, len(pr.stepsList))), [len(pr.stepsList)]]
    pr.removeInputsFlags = [False, mergeRemoveInputsFlag]
    pr.priorities = [strippPriority, mergingPriority]
    pr.cpus = [strippCPU, mergingCPU]
    pr.groupSizes = [strippFilesPerJob, mergingGroupSize]
    pr.plugins = [strippPlugin, mergingPlugin]
    pr.inputs = [strippInputDataList, []]
    pr.inputDataPolicies = [strippIDPolicy, mergingIDPolicy]
    pr.bkQueries = ["Full", "fromPreviousProd"]
    pr.targets = [targetSite, targetSite]
    pr.multicore = [strippMulticoreFlag, mergeMulticoreFlag]
    pr.outputModes = ["Run", "Run"]
    pr.ancestorDepths = [strippAncestorDepth, 0]
    pr.compressionLvl = [strippCompressionLvl] * len(pr.stepsInProds[0]) + [mergeCompressionLvl] * len(
        pr.stepsInProds[1]
    )
    pr.outputVisFlag = [{"1": strippOutputVisFlag}, {"2": mergeOutputVisFlag}]
    pr.specialOutputVisFlag = [{"1": strippOutputVisFlagSpecial}, {"2": mergeOutputVisFlagSpecial}]

elif w3:
    pr.prodsTypeList = [recoType, "Merge"]
    pr.outputSEs = [recoDataSE, mergingDataSE]
    pr.specialOutputSEs = [recoDataSESpecial, mergingDataSESpecial]
    pr.stepsInProds = [list(range(1, len(pr.stepsList))), [len(pr.stepsList)]]
    pr.removeInputsFlags = [False, mergeRemoveInputsFlag]
    pr.priorities = [recoPriority, mergingPriority]
    pr.cpus = [recoCPU, mergingCPU]
    pr.groupSizes = [recoFilesPerJob, mergingGroupSize]
    pr.plugins = [recoPlugin, mergingPlugin]
    pr.inputs = [recoInputDataList, []]
    pr.inputDataPolicies = [recoIDPolicy, mergingIDPolicy]
    pr.bkQueries = ["Full", "fromPreviousProd"]
    pr.targets = [targetSite, targetSite]
    pr.multicore = [recoMulticoreFlag, mergeMulticoreFlag]
    pr.outputModes = ["Run", "Run"]
    pr.ancestorDepths = [recoAncestorDepth, 0]
    pr.compressionLvl = [recoCompressionLvl] * len(pr.stepsInProds[0]) + [mergeCompressionLvl] * len(pr.stepsInProds[1])
    pr.outputVisFlag = [{"1": recoOutputVisFlag}, {"2": mergeOutputVisFlag}]
    pr.specialOutputVisFlag = [{"1": recoOutputVisFlagSpecial}, {"2": mergeOutputVisFlagSpecial}]

elif w4:
    pr.prodsTypeList = [recoType, "DataStripping", "Merge"]
    pr.outputSEs = [recoDataSE, strippDataSE, mergingDataSE]
    pr.specialOutputSEs = [recoDataSESpecial, strippDataSESpecial, mergingDataSESpecial]
    pr.stepsInProds = [
        list(range(1, len(pr.stepsList) - 1)),
        list(range(len(pr.stepsList) - 1, len(pr.stepsList))),
        [len(pr.stepsList)],
    ]
    pr.removeInputsFlags = [False, False, mergeRemoveInputsFlag]
    pr.priorities = [recoPriority, strippPriority, mergingPriority]
    pr.cpus = [recoCPU, strippCPU, mergingCPU]
    pr.groupSizes = [recoFilesPerJob, strippFilesPerJob, mergingGroupSize]
    pr.plugins = [recoPlugin, strippPlugin, mergingPlugin]
    pr.inputs = [recoInputDataList, [], []]
    pr.inputDataPolicies = [recoIDPolicy, strippIDPolicy, mergingIDPolicy]
    pr.bkQueries = ["Full", "fromPreviousProd", "fromPreviousProd"]
    pr.targets = [targetSite, targetSite, targetSite]
    pr.multicore = [recoMulticoreFlag, strippMulticoreFlag, mergeMulticoreFlag]
    pr.outputModes = ["Run", "Run", "Run"]
    pr.ancestorDepths = [recoAncestorDepth, strippAncestorDepth, 0]
    pr.compressionLvl = (
        [recoCompressionLvl] * len(pr.stepsInProds[0])
        + [strippCompressionLvl] * len(pr.stepsInProds[1])
        + [mergeCompressionLvl] * len(pr.stepsInProds[2])
    )
    pr.outputVisFlag = [{"1": recoOutputVisFlag}, {"2": strippOutputVisFlag}, {"3": mergeOutputVisFlag}]
    pr.specialOutputVisFlag = [
        {"1": recoOutputVisFlagSpecial},
        {"2": strippOutputVisFlagSpecial},
        {"3": mergeOutputVisFlagSpecial},
    ]

elif w5:
    pr.prodsTypeList = ["DataStripping", "Merge", "WGProduction"]
    pr.outputSEs = [strippDataSE, mergingDataSE, indexingDataSE]
    pr.specialOutputSEs = [strippDataSESpecial, mergingDataSESpecial, {}]
    pr.stepsInProds = [
        list(range(1, len(pr.stepsList) - 1)),
        list(range(len(pr.stepsList) - 1, len(pr.stepsList))),
        [len(pr.stepsList)],
    ]
    pr.removeInputsFlags = [False, mergeRemoveInputsFlag, False]
    pr.priorities = [strippPriority, mergingPriority, indexingPriority]
    pr.cpus = [strippCPU, mergingCPU, indexingCPU]
    pr.groupSizes = [strippFilesPerJob, mergingGroupSize, indexingGroupSize]
    pr.plugins = [strippPlugin, mergingPlugin, indexingPlugin]
    pr.inputs = [strippInputDataList, [], []]
    pr.inputDataPolicies = [strippIDPolicy, mergingIDPolicy, indexingIDPolicy]
    pr.bkQueries = ["Full", "fromPreviousProd", "fromPreviousProd"]
    pr.targets = [targetSite, targetSite, targetSite]
    pr.multicore = [strippMulticoreFlag, mergeMulticoreFlag, False]
    pr.outputModes = ["Run", "Run", "Any"]
    pr.ancestorDepths = [strippAncestorDepth, 0, 0]
    pr.compressionLvl = (
        [strippCompressionLvl] * len(pr.stepsInProds[0])
        + [mergeCompressionLvl] * len(pr.stepsInProds[1])
        + [""] * len(pr.stepsInProds[2])
    )
    pr.outputVisFlag = [{"1": strippOutputVisFlag}, {"2": mergeOutputVisFlag}]
    pr.specialOutputVisFlag = [{"1": strippOutputVisFlagSpecial}, {"2": mergeOutputVisFlagSpecial}]


pr.buildAndLaunchRequest()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File

from DIRAC.TransformationSystem.Client.BodyPlugin.BaseBody import BaseBody


class RAWReplicationBody(BaseBody):
    """
    This body is to be used for the RAW replication of the full stream:

    * The file is originally in CERN-DAQ-EXPORT
    * We replicate it to CERN-RAW (waiting for the file to be archived)
    * We then replicate to the chosen Tier 1 buffer (CPUforRAW shares).
    * If the chosen tape (RAW shares) is on the same site as the buffer
      (80% of the time), we replicate it from the buffer.
      Otherwise we can do it in parallel, from CERN-DAQ-EXPORT
    * Finally, we remove the replicas
    """

    # This is needed to know how to serialize the object
    _attrToSerialize = []

    @staticmethod
    def _createOperation(operationType, lfns, targetSE, sourceSE=None):
        """
        Generate an Operation object of the given Type, with the
        specified targetSE and sourceSE. Associate to it File objects
        from the lfns list

        :param str operationType: Type of the operation (RemoveReplica, ReplicateAndRegister, etc)
        :param list lfns: list of LFNs on which to perform the operation
        :param str targetSE: SE name(s) targeted by the ops
        :param str sourceSE: SE name of the source
        """
        newOp = Operation()
        newOp.Type = operationType
        newOp.TargetSE = targetSE
        if sourceSE:
            newOp.SourceSE = sourceSE
        for lfn in lfns:
            opFile = File()
            opFile.LFN = lfn
            newOp.addFile(opFile)
        return newOp

    def taskToRequest(self, taskID, task, transID):
        """
        Create the request object from the task Dict


        """

        if isinstance(task["InputData"], list):
            lfns = task["InputData"]
        elif isinstance(task["InputData"], str):
            lfns = task["InputData"].split(";")
        else:
            raise ValueError(f"InputData must be a list or a string, not {type(task['InputData'])}")

        req = Request()

        # Replicate first to CERN-RAW
        repCernRAW = self._createOperation("ReplicateAndRegister", lfns, "CERN-RAW")
        req.addOperation(repCernRAW)

        # Now look at the target SEs
        # We expect one "Site-BUFFER" SE and one "Site-RAW" SE
        targetSEs = task["TargetSE"].split(",")

        # If we have two targetSEs
        # (we should always so far, maybe not for turbo)
        if len(targetSEs) == 2:
            # Order them to have the buffer first
            bufferSE, rawSE = sorted(targetSEs, key=lambda se: "-BUFFER" in se, reverse=True)

            # Here, we can just rely on naming convention, and the fact
            # that the site always appears first in the SE names

            diskAndTapeSameSite = bufferSE.split("-BUFFER")[0] == rawSE.split("-RAW")[0]

            # If the buffer and raw SEs are on the same site,
            # do the transfer in two steps to maximise LAN transfer
            # over WAN
            if diskAndTapeSameSite:
                # First replicate to disk
                repBuffer = self._createOperation("ReplicateAndRegister", lfns, bufferSE)
                req.addOperation(repBuffer)

                # Then replicate to tape from disk
                repTape = self._createOperation("ReplicateAndRegister", lfns, rawSE, sourceSE=bufferSE)
                req.addOperation(repTape)

            # Otherwise, do both replication from EOS
            else:
                # One operation for both replication
                repBufferAndTape = self._createOperation("ReplicateAndRegister", lfns, task["TargetSE"])
                req.addOperation(repBufferAndTape)

        # If there is only one SE (like for the Turbo stream?) or more than 2 (?!)
        # do only one operation
        else:
            repBufferAndTape = self._createOperation("ReplicateAndRegister", lfns, task["TargetSE"])
            req.addOperation(repBufferAndTape)

        # Finally, remove the replicas from CERN-DAQ-EXPORT
        daqExportRemoval = self._createOperation("RemoveReplica", lfns, "CERN-DAQ-EXPORT")
        req.addOperation(daqExportRemoval)

        return req



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""RAWIntegrityDB is the front end for the database containing the files which
are awating migration.

It offers a simple interface to add files, get files and modify their
status.
"""
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities.List import stringListToString

gLogger.initialize("DMS", "/Databases/RAWIntegrityDB/Test")


class RAWIntegrityDB(DB):
    _tablesDict = {}
    # Files table
    _tablesDict["Files"] = {
        "Fields": {
            "ID": "BIGINT NOT NULL AUTO_INCREMENT",
            "LFN": "VARCHAR(255) NOT NULL",
            "PFN": "VARCHAR(255) NOT NULL",
            "Size": "BIGINT NOT NULL",
            "StorageElement": "VARCHAR(32) NOT NULL",
            "GUID": "VARCHAR(255) NOT NULL",
            "FileChecksum": "VARCHAR(255) NOT NULL",
            "SubmitTime": "DATETIME NOT NULL",
            "CompleteTime": "DATETIME",
            "Status": 'VARCHAR(255) DEFAULT "Active"',
        },
        "PrimaryKey": "ID",
        "Indexes": {"Status": ["Status"], "LFN": ["LFN"]},
    }
    # LastMonitor table
    _tablesDict["LastMonitor"] = {"Fields": {"LastMonitorTime": "DATETIME NOT NULL PRIMARY KEY"}}

    def __init__(self, systemInstance="Default"):
        DB.__init__(self, "RAWIntegrityDB", "DataManagement/RAWIntegrityDB")

    def _checkTable(self):
        """_checkTable.

        Make sure the table is created
        """

        return self.__createTables()

    def __createTables(self):
        """__createTables.

        Writes the schema in the database. If a table is already in the
        schema, it is skipped to avoid problems trying to create a table
        that already exists.
        """

        # Horrible SQL here !!
        existingTables = self._query("show tables")
        if not existingTables["OK"]:
            return existingTables
        existingTables = [existingTable[0] for existingTable in existingTables["Value"]]

        # Makes a copy of the dictionary _tablesDict
        tables = {}
        tables.update(self._tablesDict)

        for existingTable in existingTables:
            if existingTable in tables:
                del tables[existingTable]

        res = self._createTables(tables)
        if not res["OK"]:
            return res

        # Human readable S_OK message
        if res["Value"] == 0:
            res["Value"] = "No tables created"
        else:
            res["Value"] = f"Tables created: {','.join(list(tables))}"

            # Create the initial value
            if "LastMonitor" in tables:
                ret = self._query("INSERT INTO LastMonitor(LastMonitorTime) value (UTC_TIMESTAMP());")
                if not ret["OK"]:
                    return ret

        return res

    def showTables(self):
        """return the list of tables."""

        existingTables = self._query("show tables")
        if not existingTables["OK"]:
            return existingTables
        existingTables = [existingTable[0] for existingTable in existingTables["Value"]]
        return S_OK(existingTables)

    def getActiveFiles(self):
        """Obtain all the active files in the database along with all their
        associated metadata."""
        try:
            gLogger.info("RAWIntegrityDB.getActiveFiles: " "Obtaining files awaiting migration from database.")
            res = self._query(
                "SELECT "
                "LFN,PFN,Size,StorageElement,GUID,FileChecksum,TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(),SubmitTime)) "
                "from Files WHERE Status = 'Active';"
            )
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.getActiveFiles: Failed to get files from database.", res["Message"])
                return res
            else:
                fileDict = {}
                for lfn, pfn, size, se, guid, checksum, waittime in res["Value"]:
                    fileDict[lfn] = {
                        "PFN": pfn,
                        "Size": size,
                        "SE": se,
                        "GUID": guid,
                        "Checksum": checksum,
                        "WaitTime": waittime,
                    }
                gLogger.info(
                    f"RAWIntegrityDB.getActiveFiles: Obtained {len(fileDict)} files awaiting migration from database."
                )
                return S_OK(fileDict)
        except Exception as x:
            errStr = "RAWIntegrityDB.getActiveFiles: Exception while getting files from database."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def getFiles(self, status):
        """Obtain all the files in a give status (or list) in the database along
        with all their associated metadata.

        :param status: string or list of status we want to query

        :returns: dict {lfn: metadata}
        """
        if isinstance(status, str):
            status = [status]

        try:
            gLogger.info(f"RAWIntegrityDB.getActiveFiles: Obtaining files in status {status}.")
            res = self._query(
                "SELECT "
                "LFN,Status,PFN,Size,StorageElement,GUID,FileChecksum,"
                "TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(),SubmitTime)),SubmitTime "
                "from Files WHERE Status in (%s);" % (stringListToString(status))
            )
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.getActiveFiles: Failed to get files from database.", res["Message"])
                return res
            fileDict = {}
            for lfn, st, pfn, size, se, guid, checksum, waittime, submitTime in res["Value"]:
                fileDict[lfn] = {
                    "PFN": pfn,
                    "Size": size,
                    "SE": se,
                    "GUID": guid,
                    "Checksum": checksum,
                    "WaitTime": waittime,
                    "Status": st,
                    "SubmitTime": submitTime,
                }
            gLogger.info(
                f"RAWIntegrityDB.getActiveFiles: Obtained {len(fileDict)} files awaiting migration from database."
            )
            return S_OK(fileDict)
        except Exception as x:
            errStr = "RAWIntegrityDB.getActiveFiles: Exception while getting files from database."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def getUnmigratedFiles(self):
        """Get files in status Active, Copied, and Registered."""

        return self.getFiles(["Active", "Copied", "Registered"])

    def setFileStatus(self, lfn, status):
        """Update the status of the file in the database.

        A file in status 'Done' cannot be unchanged
        """
        try:
            gLogger.info(f"RAWIntegrityDB.setFileStatus: Attempting to update status of {lfn} to '{status}'.")
            req = "UPDATE Files SET Status='{}',CompleteTime=UTC_TIMESTAMP() WHERE LFN = '{}' AND Status != 'Done';".format(
                status,
                lfn,
            )
            res = self._update(req)
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.setFileStatus: Failed update file status.", res["Message"])
            else:
                gLogger.info("RAWIntegrityDB.setFileStatus: Successfully updated file status.")
            return res
        except Exception as x:
            errStr = "RAWIntegrityDB.setFileStatus: Exception while updating file status."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def removeFile(self, lfn):
        """Remove file from the DB."""
        try:
            gLogger.info(f"RAWIntegrityDB.removeFile: Attempting to remove {lfn}.")
            req = f"DELETE FROM Files WHERE LFN = '{lfn}';"
            res = self._update(req)
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.removeFile: Failed to remove file.", res["Message"])
            else:
                gLogger.info("RAWIntegrityDB.removeFile: Successfully removed file.")
            return res
        except Exception as x:
            errStr = "RAWIntegrityDB.removeFile: Exception while removing file."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def addFile(self, lfn, pfn, size, se, guid, checksum):
        """Insert file into the database."""
        try:
            gLogger.info(f"RAWIntegrityDB.addFile: Attempting to add {lfn} to database.")
            req = (
                "INSERT INTO Files (LFN,PFN,Size,StorageElement,GUID,FileChecksum,SubmitTime) VALUES\
            ('%s','%s',%s,'%s','%s','%s',UTC_TIMESTAMP());"
                % (lfn, pfn, size, se, guid, checksum)
            )
            res = self._update(req)
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.addFile: Failed update add file to database.", res["Message"])
            else:
                gLogger.info("RAWIntegrityDB.addFile: Successfully added file.")
            return res
        except Exception as x:
            errStr = "RAWIntegrityDB.addFile: Exception while updating file status."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def setLastMonitorTime(self):
        """Set the last time the migration rate was calculated."""
        try:
            gLogger.info("RAWIntegrityDB.setLastMonitorTime: Attempting to set the last migration marker.")
            req = "UPDATE LastMonitor SET LastMonitorTime=UTC_TIMESTAMP();"
            res = self._update(req)
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.setLastMonitorTime: Failed update migration marker.", res["Message"])
            else:
                gLogger.info("RAWIntegrityDB.setLastMonitorTime: Successfully updated migration marker.")
            return res
        except Exception as x:
            errStr = "RAWIntegrityDB.setLastMonitorTime: Exception while updating migration marker."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def getLastMonitorTimeDiff(self):
        """Get the last time the migration rate was calculated."""
        try:
            gLogger.info("RAWIntegrityDB.getLastMonitorTimeDiff: Attempting to get the last migration marker.")
            req = "SELECT TIME_TO_SEC(TIMEDIFF(UTC_TIMESTAMP(),LastMonitorTime)) FROM LastMonitor LIMIT 1;"
            res = self._query(req)
            if not res["OK"]:
                gLogger.error("RAWIntegrityDB.getLastMonitorTimeDiff: Failed get migration marker.", res["Message"])
                return res
            else:
                gLogger.info("RAWIntegrityDB.getLastMonitorTimeDiff: Successfully obtained migration marker.")
                timediff = res["Value"][0][0]
                return S_OK(timediff)
        except Exception as x:
            errStr = "RAWIntegrityDB.getLastMonitorTimeDiff: Exception while getting migration marker."
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

    def getGlobalStatistics(self):
        """Get the count of the file statutes in the DB."""
        req = "SELECT Status,COUNT(*) FROM Files GROUP BY Status;"
        res = self._query(req)
        if not res["OK"]:
            return res
        statusDict = {}
        for resValue in res["Value"]:
            status, count = resValue
            statusDict[status] = count
        return S_OK(statusDict)

    def getFileSelections(self):
        """Get the unique values of the selection fields."""
        selDict = {"StorageElement": [], "Status": []}
        req = "SELECT DISTINCT(StorageElement) FROM Files;"
        res = self._query(req)
        if not res["OK"]:
            return res
        for resValue in res["Value"]:
            selDict["StorageElement"].append(resValue[0])
        req = "SELECT DISTINCT(Status) FROM Files;"
        res = self._query(req)
        if not res["OK"]:
            return res
        for resValue in res["Value"]:
            selDict["Status"].append(resValue[0])
        return S_OK(selDict)

    def __buildCondition(self, condDict, older=None, newer=None, timeStamp="SubmitTime"):
        """build SQL condition statement from provided condDict and other extra
        conditions."""
        condition = ""
        conjunction = "WHERE"
        if condDict is not None:
            for attrName, attrValue in condDict.items():
                ret = self._escapeString(attrName)
                if not ret["OK"]:
                    return ret
                attrName = "`" + ret["Value"][1:-1] + "`"
                if isinstance(attrValue, list):
                    multiValueList = []
                    for x in attrValue:
                        ret = self._escapeString(x)
                        if not ret["OK"]:
                            return ret
                        x = ret["Value"]
                        multiValueList.append(x)
                    multiValue = ",".join(multiValueList)
                    condition = f" {condition} {conjunction} {attrName} in ({multiValue})"
                else:
                    ret = self._escapeString(attrValue)
                    if not ret["OK"]:
                        return ret
                    attrValue = ret["Value"]
                    condition = f" {condition} {conjunction} {attrName}={attrValue}"
                conjunction = "AND"
        if older:
            ret = self._escapeString(older)
            if not ret["OK"]:
                return ret
            older = ret["Value"]
            condition = f" {condition} {conjunction} {timeStamp} <= {older}"
            conjunction = "AND"
        if newer:
            ret = self._escapeString(newer)
            if not ret["OK"]:
                return ret
            newer = ret["Value"]
            condition = f" {condition} {conjunction} {timeStamp} >= {newer}"
        return condition

    def selectFiles(self, selectDict, orderAttribute="LFN", newer=None, older=None, limit=None):
        """Select the files which match the selection criteria."""
        condition = self.__buildCondition(selectDict, older, newer)
        if orderAttribute:
            orderType = None
            orderField = orderAttribute
            if orderAttribute.find(":") != -1:
                orderType = orderAttribute.split(":")[1].upper()
                orderField = orderAttribute.split(":")[0]
            condition = condition + " ORDER BY " + orderField
            if orderType:
                condition = condition + " " + orderType
        if limit:
            condition = condition + " LIMIT " + str(limit)
        res = self._query(
            "SELECT "
            "LFN,PFN,Size,StorageElement,GUID,FileChecksum,SubmitTime,CompleteTime,Status "
            "from Files %s" % condition
        )
        if not res["OK"]:
            return res
        if not len(res["Value"]):
            return S_OK([])
        return S_OK(res["Value"])


# ...............................................................................



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Create and send a combined request for any pending operations at the end of
a job:

fileReport (for the transformation) jobReport (for jobs) accounting
request (for failover)
"""
from DIRAC import S_OK, S_ERROR, gLogger

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class FailoverRequest(ModuleBase):
    #############################################################################

    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""

        self.log = gLogger.getSubLogger("FailoverRequest")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here."""
        super()._resolveInputVariables()
        super()._resolveInputStep()

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution function."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            if not self._enableModule():
                return S_OK()

            self._resolveInputVariables()

            # preparing the request, just in case
            self.request.RequestName = "job_%d_request.xml" % self.jobID
            self.request.JobID = self.jobID
            self.request.SourceComponent = "Job_%d" % self.jobID

            # report on the status of the input data, by default they are 'Processed', unless the job failed
            # failures happening before (e.g. in previous steps, or while inspecting the XML summary) are not touched.

            # The FileReport object is normally empty, unless there are some Problematic files,
            # or if there are files found to have descendants
            filesInFileReport = self.fileReport.getFiles()

            if not self._checkWFAndStepStatus(noPrint=True):
                # if the job fails, the input files will be reset to 'Unused'
                for lfn in self.inputDataList:
                    if lfn not in filesInFileReport:
                        self.log.info("Set status of input files to 'Unused' due to workflow failure", lfn)
                        self.fileReport.setFileStatus(int(self.production_id), lfn, "Unused")
            else:
                for lfn in self.inputDataList:
                    if lfn not in filesInFileReport:
                        self.log.verbose("No status populated for input data, setting to 'Processed'", lfn)
                        self.fileReport.setFileStatus(int(self.production_id), lfn, "Processed")

            res_fileReportCommit = self.fileReport.commit()
            # If there are still files to set, try a second time and generate a request if it fails again
            if not res_fileReportCommit["OK"]:
                self.log.error("Something went wrong trying fileReport.commit()", res_fileReportCommit["Message"])
            else:
                if res_fileReportCommit["Value"]:
                    self.log.info("Status of files have been properly updated in the TransformationDB")
                else:
                    self.log.warn("No file status update reported. There are no input files?")

            if self.fileReport.getFiles():
                self.log.error("On first attempt, failed to report file status to TransformationDB")
                # This will try a second time a commit, before generating a SetFileStatus operation
                result = self.fileReport.generateForwardDISET()
                if not result["OK"]:
                    self.log.warn("Could not generate Operation for file report", f"with result:\n{result['Message']}")
                else:
                    if result["Value"] is None:  # Means the FileReport managed to report, no need for a new operation
                        self.log.info("On second attempt, files correctly reported to TransformationDB")
                    else:
                        self.log.error("On second attempt, SetFileStatus definitely failed")
                        if self.workflowStatus["OK"] and self.stepStatus["OK"]:  # the job will be in "Completing"
                            self.log.info("Adding a SetFileStatus operation to the request")
                            self.request.addOperation(result["Value"])
                        else:
                            self.log.info("The job should fail: do not set requests, as the DRA will take care")

            # Must ensure that the local job report instance is used to report the final status
            # in case of failure and a subsequent failover operation
            if self.workflowStatus["OK"] and self.stepStatus["OK"]:
                self.setApplicationStatus("Job Finished Successfully")

            self.generateFailoverFile()

            return S_OK()

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in FailoverRequest execute module", lException=e)
            self.setApplicationStatus(repr(e))
            return S_ERROR(str(e))

        finally:
            super().finalize()


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""PluginScript module holds PluginScript class, which is an extension of
DMScript class."""
from DIRAC import S_OK, gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import resolveSEGroup

from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript


class Setter:
    """Class used for setting an option: use setOption() method as a script
    setter."""

    def __init__(self, obj, name):
        self.name = name
        self.obj = obj

    def setOption(self, val):
        """Method used as a setter in the switch definition."""
        if self.name.endswith("="):
            try:
                self.obj.options[self.name[:-1]] = val if not self.name == "GroupSize=" else float(val)
            except ValueError as e:
                gLogger.exception("Bad value for parameter", self.name, lException=e)
        else:
            self.obj.options[self.name] = True

        return S_OK()


class PluginScript(DMScript):
    """Scripts utilities class."""

    def __init__(self):
        super().__init__()
        self.pluginParameters = {
            "Plugin=": "   Plugin name (mandatory)",
            "Type=": "   Transformation type [Replication] (Removal automatic)",
            "Parameters=": "   Additional plugin parameters ({<key>:<val>,[<key>:val>]}",
            "RequestID=": "   Sets the request ID (default 0)",
        }
        self.seParameters = (
            "KeepSEs",
            "ArchiveSEs",
            "MandatorySEs",
            "SecondarySEs",
            "DestinationSEs",
            "FromSEs",
            "RAWStorageElements",
            "ProcessingStorageElements",
        )
        self.additionalParameters = {
            "GroupSize=": "   GroupSize parameter for merging (GB) or nb of files",
            "NumberOfReplicas=": "   Number of copies to create or to remove",
            "ProcessingPasses=": "   List of processing passes for the DeleteReplicasWhenProcessed plugin",
            "Period=": "   minimal period at which a plugin is executed (if instrumented)",
            "CleanTransformations": "   (only for DestroyDataset) clean transformations from the files being destroyed",
            "Debug": "   Sets a debug flag in the plugin",
            "UseRunDestination": "   for RAWReplication plugin, use the already defined run destination as storage",
            "ThrottlePendingTasks=": "   throttle the number of Waiting jobs at each TargetSE",
        }
        self.setters = {}

    def registerPluginSEParameters(self):
        for param in self.seParameters:
            param += "="
            self.setters[param] = Setter(self, param)
            Script.registerSwitch(
                "", param, "   List of SEs for the corresponding parameter of the plugin", self.setters[param].setOption
            )

    def registerPluginAdditionalParameters(self):
        for option in self.additionalParameters:
            self.setters[option] = Setter(self, option)
            Script.registerSwitch("", option, self.additionalParameters[option], self.setters[option].setOption)

    def registerPluginSwitches(self):
        """Set of switches used by TS plugins."""
        self.registerBKSwitches()

        for option in self.pluginParameters:
            self.setters[option] = Setter(self, option)
            Script.registerSwitch("", option, self.pluginParameters[option], self.setters[option].setOption)

        self.registerPluginSEParameters()
        self.registerPluginAdditionalParameters()

    def getPluginParameters(self):
        """Get  parameters used by TS plugins."""
        if "Parameters" in self.options:
            params = eval(self.options["Parameters"])  # pylint: disable=eval-used
        else:
            params = {}
        # print self.options
        for key in set(self.options) & {
            param if not param.endswith("=") else param[:-1] for param in self.additionalParameters
        }:
            params[key] = self.options[key]
        return params

    def getPluginSEParameters(self):
        """Special treatment for SE-related parameters."""
        params = {}
        # print self.options
        for key in set(self.options) & set(self.seParameters):
            val = self.options[key]
            params[key] = val if val else []
        return params



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Module GaudiApplication."""
import six

import copy
import os

from unittest.mock import MagicMock
import pytest

pytestmark = pytest.mark.skipif(six.PY2, reason="This test only supports Python 3")

from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
)


@pytest.fixture
def rmFiles():
    yield
    try:
        os.remove("gaudi_extra_options.py")
    except OSError:
        pass


def test_execute(mocker, rmFiles):
    from LHCbDIRAC.Workflow.Modules.GaudiApplication import GaudiApplication

    mocker.patch("LHCbDIRAC.Workflow.Modules.GaudiApplication.RunApplication", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.GaudiApplication.ModuleBase._manageAppOutput", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.GaudiApplication.gConfig", side_effect=MagicMock())
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())

    ga = GaudiApplication(bkClient=bkc_mock, dm=dm_mock)
    ga.siteName = "LCG.PIPPO.org"
    ga.jobType = "user"

    # no errors, no input data
    for wf_cs in copy.deepcopy(wf_commons):
        for s_cs in step_commons:
            assert (
                ga.execute(
                    prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                )["OK"]
                is True
            )



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Get production numbers given a dataset path."""
import time
from DIRAC import gLogger, exit
from DIRAC.Core.Base.Script import Script


def printProds(title, prods):
    typeDict = {}
    for prod, prodType in prods.items():
        typeDict.setdefault(prodType, []).append(prod)
    gLogger.notice(title)
    for prodType, prodList in typeDict.items():
        gLogger.notice(f"({prodType}): {','.join([str(prod) for prod in sorted(prodList)])}")


def execute(dmScript):
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    tr = TransformationClient()

    for switch in Script.getUnprocessedSwitches():
        pass

    bkQuery = dmScript.getBKQuery()
    if not bkQuery:
        gLogger.notice("No BKQuery given...")
        exit(1)

    startTime = time.time()
    prods = bkQuery.getBKProductions()  # visible = 'All' )

    parents = {}
    productions = {}
    for prod in prods:
        ptype = tr.getTransformation(prod).get("Value", {}).get("Type", "Unknown")
        productions[prod] = ptype
        parent = tr.getBookkeepingQuery(prod).get("Value", {}).get("ProductionID", "")
        while isinstance(parent, int):
            ptype = tr.getTransformation(parent).get("Value", {}).get("Type", "Unknown")
            parents[parent] = ptype
            parent = tr.getBookkeepingQuery(parent).get("Value", {}).get("ProductionID", "")

    gLogger.notice(f"For BK path {bkQuery.getPath()}:")
    if not prods:
        gLogger.notice("No productions found!")
    else:
        printProds("Productions found", productions)
        if parents:
            printProds("Parent productions", parents)

    gLogger.notice(f"Completed in {time.time() - startTime:.1f} seconds")


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)
    execute(dmScript)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
import sys
from time import time

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC import exit, gLogger

    Script.registerSwitch("", "User=", "  User name")
    Script.registerSwitch("", "Recursive", "  Set ownership recursively")
    Script.registerSwitch("", "Directory=", "   Directory to change the ownership of")
    Script.registerSwitch("", "Group=", "   Group name (default lhcb_user)")
    Script.registerSwitch("", "Mode=", "   Change permission mode (default: not changed)")
    Script.registerSwitch("", "Create", "   Use for creating the directory if it does not exist")

    Script.parseCommandLine()

    user = None
    recursive = False
    group = None
    dirList = None
    mode = None
    create = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "User":
            user = switch[1].lower()
        elif switch[0] == "Recursive":
            recursive = True
        elif switch[0] == "Directory":
            dirList = switch[1].split(",")
        elif switch[0] == "Group":
            group = switch[1]
        elif switch[0] == "Mode":
            mode = int(switch[1], base=8)
        elif switch[0] == "Create":
            create = True

    args = Script.getPositionalArgs()
    if dirList is None and args:
        dirList = []
        for arg in args:
            dirList += arg.split(",")

    toChange = []
    if user:
        toChange.append("user")
    if group:
        toChange.append("group")
    if mode:
        toChange.append("mode")
    toChange = ",".join(toChange)

    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    from LHCbDIRAC.DataManagementSystem.Utilities.FCUtilities import chown

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.fatal("Can't get proxy info", res["Message"])
        exit(1)
    properties = res["Value"].get("groupProperties", [])
    if "FileCatalogManagement" not in properties:
        gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
        exit(5)

    if group and not group.startswith("lhcb_"):
        gLogger.fatal("This is not a valid group", group)
        exit(1)

    directories = set()
    for baseDir in dirList:
        if os.path.exists(baseDir):
            f = open(baseDir)
            directories.update([os.path.dirname(line.split()[0] + "/") for line in f.read().splitlines()])
            f.close()
        else:
            directories.add(baseDir if not baseDir.endswith("/") else baseDir[:-1])

    if not directories or (user is None and group is None and mode is None):
        Script.showHelp(exitCode=1)

    from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar
    from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

    dfc = FileCatalogClient()

    success = 0
    startTime = time()
    progressBar = ProgressBar(
        len(directories),
        title=f"Changing {toChange}{' recursively' if recursive else ''} to {len(directories)} directories:",
        chunk=1,
        interactive=True,
    )
    errors = {}
    for baseDir in directories:
        progressBar.loop()
        if not baseDir.startswith("/lhcb"):
            errors[baseDir] = "Not a valid directory"
            continue
        if not dfc.isDirectory(baseDir).get("Value", {}).get("Successful", {}).get(baseDir):
            if create:
                res = dfc.createDirectory(baseDir)
                if not res["OK"]:
                    errors[baseDir] = res["Message"] + " while creating directory"
                    continue
            else:
                errors[baseDir] = "Directory doesn't exist"
                continue
        res = chown(baseDir, user, group=group, mode=mode, recursive=recursive, fcClient=dfc)
        if not res["OK"]:
            errors[baseDir] = res["Message"] + f" executing {res['Action']}: "
        else:
            success += res["Value"]

    if success:
        msg = "Successfully changed %s%s in %d directories" % (toChange, " recursively" if recursive else "", success)
    else:
        msg = "Failed changing %s%s in %d directories" % (
            toChange,
            " recursively" if recursive else "",
            len(directories),
        )
    progressBar.endLoop(msg)
    retCode = 0
    if errors:
        retCode = 1
        gLogger.notice("Errors:")
        for baseDir, error in errors.items():
            gLogger.notice(f"\tDirectory {baseDir} - ", error)
    elif len(directories) == 1:
        from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript
        from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeLfnMetadata

        dmScript = DMScript()
        dmScript.setLFNs(list(directories)[0])
        sys.stdout.write("Directory metadata: ")
        sys.stdout.flush()
        executeLfnMetadata(dmScript)
    exit(retCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Module containing a front-end to the ElasticSearch-based ElasticLogErrorsDB.
    This module interacts with one ES index: "ElasticMCLogErrors",

    Here we define a mapping which is taken from a list of log errors.
"""

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.ElasticDB import ElasticDB
from DIRAC.Core.Utilities.ReturnValues import DReturnType


class ElasticMCStatsDBBase(ElasticDB):
    def set(self, data: list[dict]) -> DReturnType:
        """
        Inserts data into ElasticDB index

        :param data: data to be inserted

        :returns: S_OK/S_ERROR as result of indexing
        """
        self.log.debug(
            self.__class__.__name__,
            f".set(): inserting data in {self.indexName}:{data}",  # pylint: disable=no-member
        )

        result = self.index(
            self.indexName,  # pylint: disable=no-member
            body=data,
            docID=str(data["ProductionID"]) + "_" + str(data["JobID"]),
        )
        if not result["OK"]:
            self.log.error("ERROR: Couldn't insert data", result["Message"])
        return result

    def get(self, productionID: int) -> DReturnType[list]:
        """Get docs per productionID. Basically here only for tests, right now

        :param productionID: production ID
        :return: list with all docs
        """
        self.log.debug(self.__class__.__name__ + f".get(): Getting for production {productionID}")

        resultList = []

        """ the following should be equivalent to
        {
          "query": {
            "bool": {
              "filter": {  # no scoring
                "term": {"ProductionID": productionID}  # term level query, does not pass through the analyzer
              }
            }
          }
        }
        """

        s = self.dslSearch.query("bool", filter=self._Q("term", ProductionID=productionID))  # pylint: disable=no-member

        res = s.execute()

        for hit in res:
            hitDict = {}
            for name in hit:
                hitDict[name] = getattr(hit, name)
            resultList.append(hitDict)
        return S_OK(resultList)

    def remove(self, productionID: int) -> DReturnType:
        """Remove docs per productionID. Basically here only for tests, right now

        :param productionID: production ID
        :return: S_OK/S_ERROR
        """

        self.log.debug(self.__class__.__name__ + f".remove(): Removing documents of production {productionID}")

        """ the following should be equivalent to
        {
          "query": {
            "bool": {
              "filter": {  # no scoring
                "term": {"ProductionID": productionID}  # term level query, does not pass through the analyzer
              }
            }
          }
        }
        """

        s = self.dslSearch.query("bool", filter=self._Q("term", ProductionID=productionID))  # pylint: disable=no-member
        try:
            s.delete()
        except Exception as e:
            self.log.exception()
            return S_ERROR(repr(e))

        return S_OK()



###############################################################################
# (c) Copyright 2023 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

from DIRAC import gLogger
from DIRAC.Core.Utilities.ReturnValues import DReturnType, returnValueOrRaise, convertToReturnValue
from .FileQueryUtils import TableJoins, buildQuery, combineDescription, TABLE_JOIN_TO_NAME


GETFILESWITHMETADATA_AVAILABLE_COLUMNS = {
    "filename": None,
    "eventstat": None,
    "filesize": None,
    "creationdate": None,
    "jobstart": TableJoins.JOBS,
    "jobend": TableJoins.JOBS,
    "workernode": TableJoins.JOBS,
    "name": TableJoins.FILETYPES,
    "runnumber": TableJoins.JOBS,
    "fillnumber": TableJoins.JOBS,
    "fullstat": None,
    "dataqualityflag": TableJoins.DATAQUALITY,
    "eventinputstat": TableJoins.JOBS,
    "totalluminosity": TableJoins.JOBS,
    "luminosity": None,
    "instluminosity": None,
    "tck": TableJoins.JOBS,
    "guid": None,
    "adler32": None,
    "eventtypeid": None,
    "md5sum": None,
    "visibilityflag": None,
    "jobid": TableJoins.JOBS,
    "gotreplica": None,
    "inserttimestamp": None,
}

GETFILESWITHMETADATA_NAME_TO_COL = {
    "FileName": "filename",
    "EventStat": "eventstat",
    "FileSize": "filesize",
    "CreationDate": "creationdate",
    "JobStart": "jobstart",
    "JobEnd": "jobend",
    "WorkerNode": "workernode",
    "FileType": "name",
    "RunNumber": "runnumber",
    "FillNumber": "fillnumber",
    "FullStat": "fullstat",
    "DataqualityFlag": "dataqualityflag",
    "EventInputStat": "eventinputstat",
    "TotalLuminosity": "totalluminosity",
    "Luminosity": "luminosity",
    "InstLuminosity": "instluminosity",
    "TCK": "tck",
    "GUID": "guid",
    "ADLER32": "adler32",
    "EventType": "eventtypeid",
    "MD5SUM": "md5sum",
    "VisibilityFlag": "visibilityflag",
    "JobId": "jobid",
    "GotReplica": "gotreplica",
    "InsertTimeStamp": "inserttimestamp",
}


class NewOracleBookkeepingDB:
    def __init__(self, *, dbW, dbR):
        self.log = gLogger.getSubLogger("LegacyOracleBookkeepingDB")
        self.dbW_ = dbW
        self.dbR_ = dbR

    def getAvailableFileTypes(self) -> DReturnType[list[str]]:
        """Retrieve all available file types from the database."""
        return self.dbR_.executeStoredProcedure("BOOKKEEPINGORACLEDB.getAvailableFileTypes", [])

    @convertToReturnValue
    def getFileTypesForProdID(self, prodID: int) -> list[str]:
        query_parts = [
            "SELECT DISTINCT filetypes.name",
            "FROM files, jobs, filetypes",
            "WHERE files.jobid = jobs.jobid AND jobs.production = :prodid AND filetypes.filetypeid = files.filetypeid",
        ]
        result = returnValueOrRaise(self.dbR_.query(" ".join(query_parts), kwparams={"prodid": prodID}))
        return [ft for ft, in result]

    @convertToReturnValue
    def getAvailableSMOG2States(self) -> list[str]:
        """Retrieve all available SMOG2 states."""
        result = returnValueOrRaise(self.dbR_.query("SELECT state FROM smog2"))
        return [state for state, in result]

    @convertToReturnValue
    def getRunsForSMOG2(self, state: str) -> list[int]:
        """Retrieve all runs with specified SMOG2 state

        :param str state: required state
        """
        query = "SELECT runs.runnumber FROM smog2 LEFT JOIN runs ON runs.smog2_id = smog2.id WHERE smog2.state = :state"
        result = returnValueOrRaise(self.dbR_.query(query, kwparams={"state": state}))
        return [run for run, in result]

    def setSMOG2State(self, state: str, update: bool, runs: list[int]) -> DReturnType[None]:
        """Set SMOG2 state for runs.

        :param str state: state for given runs
        :param bool update: when True, updates existing state, when False throw an error in such case
        :param list[int] runs: runs list
        """
        return self.dbW_.executeStoredProcedure(
            "BOOKKEEPINGORACLEDB.setSMOG2", parameters=[state, update], output=False, array=runs
        )

    def setExtendedDQOK(self, run: int, update: bool, dqok: list[str]) -> DReturnType[None]:
        """Set ExtendedDQOK for specified run and systems. In case update is allowed,
        not specified systems are unset for the run.

        :param int run: run number for which systems are specified
        :param bool update: when True, updates existing set, when False throw an error in such case
        :param list[str] dqok: list of system names
        """
        return self.dbW_.executeStoredProcedure(
            "BOOKKEEPINGORACLEDB.setExtendedDQOK", parameters=[run, update, dqok], output=False
        )

    @convertToReturnValue
    def getRunsWithExtendedDQOK(self, dqok: list[str]) -> list[int]:
        """Retrieve all runs with specified systems in ExtendedDQOK
        NOTE: it is NOT checking quality is set to OK, so it should NOT be used
        for end user operations.

        :param list[str] dqok: systems
        """
        if not dqok:
            return []
        sql = ["SELECT ok.runnumber FROM extendeddqok ok"]
        params = {"sysname0": dqok[0]}
        for i, system in enumerate(dqok[1::]):
            sql.append(
                f"INNER JOIN extendeddqok ok{i} ON ok{i}.runnumber = ok.runnumber AND ok{i}.systemname = :sysname{i}"
            )
            params[f"sysname{i}"] = system
        sql.append("WHERE ok.systemname = :sysname0")
        result = returnValueOrRaise(self.dbR_.query(" ".join(sql), kwparams=params))
        return [run for run, in result]

    @convertToReturnValue
    def getRunExtendedDQOK(self, runnb: int) -> list[str]:
        """Return the list of systems in ExtendedDQOK for given run

        :param int runnb: run number
        """
        query = "SELECT systemname FROM extendeddqok WHERE runnumber = :run"
        result = returnValueOrRaise(self.dbR_.query(query, kwparams={"run": runnb}))
        return [sysname for sysname, in result]

    @convertToReturnValue
    def getAvailableExtendedDQOK(self) -> list[str]:
        """Retrieve all available Extended DQOK systems."""
        result = returnValueOrRaise(self.dbR_.query("select distinct systemname from extendeddqok"))
        return [systemname for systemname, in result]

    @convertToReturnValue
    def getListOfRunsInProd(self, prod_id) -> list[int]:
        query = "SELECT DISTINCT j.runnumber FROM jobs j WHERE j.production = :prod_id"
        result = returnValueOrRaise(self.dbR_.query(query, kwparams={"prod_id": prod_id}))
        return [run for run, in result]

    @convertToReturnValue
    def getInputOutputFilesForProd(self, prod_id, run_number):
        query_parts = [
            "SELECT f2.filename aname, f.filename dname, f.gotreplica, dt.name dtype",
            "FROM jobs j",
            "LEFT JOIN files f ON j.jobid = f.jobid",
            "LEFT JOIN filetypes dt ON dt.filetypeid = f.filetypeid",
            "LEFT JOIN inputfiles i ON i.jobid = j.jobid",
            "LEFT JOIN files f2 ON i.fileid = f2.fileid",
            "where j.production = :prod_id and j.runnumber = :run_number",
        ]
        result = returnValueOrRaise(
            self.dbR_.query("\n".join(query_parts), kwparams={"prod_id": prod_id, "run_number": run_number})
        )
        return {
            "Records": result,
            "ParameterNames": ["AncestorLFN", "DescendantLFN", "GotReplica", "FileType"],
            "TotalRecords": len(result),
        }

    @convertToReturnValue
    def getOutputDescendantsForProd(self, prod_id, run_number):
        query_parts = [
            "SELECT f.filename AS aname,",
            "    f2.filename AS dname,",
            "    f2.gotreplica,",
            "    ft2.name",
            "FROM jobs j",
            "LEFT JOIN files f ON j.jobid = f.jobid",
            "INNER JOIN inputfiles i ON i.fileid = f.fileid",
            "LEFT JOIN jobs j2 ON j2.jobid = i.jobid",
            "LEFT JOIN files f2 ON j2.jobid = f2.jobid",
            "LEFT JOIN filetypes ft2 ON ft2.filetypeid = f2.filetypeid",
            "WHERE j.production = :prod_id",
            "AND j.runnumber = :run_number",
        ]
        result = returnValueOrRaise(
            self.dbR_.query("\n".join(query_parts), kwparams={"prod_id": prod_id, "run_number": run_number})
        )
        return {
            "Records": result,
            "ParameterNames": ["AncestorLFN", "DescendantLFN", "GotReplica", "FileType"],
            "TotalRecords": len(result),
        }

    def _collectFileDescendents(
        self, lfns: list[str], depth: int, production: int, checkreplica: bool, tree: bool
    ) -> dict:
        """collects descendents for specified ancestors

        :param list lfns: a list of LFNs (ancestors)
        :param int depth: the depth of the processing pass chain(how far to go)
        :param productin: production number, when not zero, search descendats in that production only
        :param bool checkreplica: when set, returned descendents should have a replica
        :param bool tree: return direct ancestors relations when true
        :returns: descendent relations and metadata
        """
        # max number of lfns to process in one query, should be less then 1000 (with current implementation)
        block_size = 100
        # allowed depth range as in original version
        depth = min(10, max(1, depth))

        processed = dict.fromkeys(lfns, False)  # to detect "NotProcessed"
        dfiles = {lfn: {} for lfn in lfns}  # requested ancestors are always included
        metadata = {}

        sql_fields = [
            f"{'PRIOR' if tree else 'CONNECT_BY_ROOT'} f.filename aname",
            "f.filename dname",
            "f.gotreplica",
            "f.eventstat",
            "f.eventtypeid",
            "f.luminosity",
            "f.instluminosity",
            "dt.name dtype",
            "j.production" if production != 0 else "0",
        ]
        sql_fields = ", ".join(sql_fields)

        sql_joins = [
            "LEFT JOIN inputfiles i ON i.fileid = f.fileid",
            "LEFT JOIN filetypes dt ON dt.filetypeid = f.filetypeid",
        ]
        if production != 0:
            sql_joins.append("LEFT JOIN jobs j ON j.jobid = f.jobid")
        sql_joins = " ".join(sql_joins)

        sql_where = f"LEVEL > 1 AND LEVEL < {depth+2}"  # LEVEL 1 is ancestor, depth is inclusive
        sql_connect = "PRIOR i.jobid = f.jobid"

        block_idx = 0
        while True:
            block_lfns = lfns[block_idx : block_idx + block_size :]
            if not block_lfns:
                break
            block_idx += block_size  # for the next block
            # bind as proposed in https://python-oracledb.readthedocs.io/en/latest/user_guide/bind.html , 7.13
            sql_start = "f.filename IN (" + ",".join([f":{i}" for i in range(1, len(block_lfns) + 1)]) + ")"

            sql = (
                f"SELECT {sql_fields} FROM files f {sql_joins} WHERE {sql_where}"
                f" START WITH {sql_start} CONNECT BY {sql_connect}"
            )
            result = returnValueOrRaise(self.dbR_.query(sql, params=[block_lfns]))
            for aname, dname, gotreplica, eventstat, eventtype, lumi, instlumi, dtype, prod in result:
                if aname in processed:  # for tree == true case
                    processed[aname] = True  # mimic previous behaviour ("processed" if there is any descendant)
                if (not checkreplica or (gotreplica != "No")) and (
                    prod == production
                ):  # should work correctly for production == 0
                    if aname not in dfiles:  # can happened when tree == true
                        dfiles[aname] = {}
                    dfiles[aname][dname] = {}
                    metadata[dname] = {
                        "GotReplica": gotreplica,
                        "EventStat": eventstat,
                        "EventType": eventtype,
                        "Luminosity": lumi,
                        "InstLuminosity": instlumi,
                        "FileType": dtype,
                    }
                elif tree:  # in case of tree, we need all relations to build it
                    if aname not in dfiles:
                        dfiles[aname] = {}
                    dfiles[aname][dname] = {}

        return processed, dfiles, metadata

    @convertToReturnValue
    def getFileDescendents(
        self, lfns: list[str], depth: int = 0, production: int = 0, checkreplica: bool = True
    ) -> dict:
        """collects descendents for specified ancestors

        :param list lfns: a list of LFNs (ancestors)
        :param int depth: the depth of the processing pass chain(how far to go)
        :param productin: production number, when not zero, search descendats in that production only
        :param bool checkreplica: when set, returned descendents should have a replica
        :returns: descendents and suplementary information
        """
        # AZ: original code never fail, it returns "Failed" list in case of problems with Oracle.
        #     So the behaviour is NOT identical when there are problems with Oracle
        # AZ: the order of files in lists of "Successful" is different from original (dictionary "WithMetadata" should match)
        if not lfns:
            return {"Failed": [], "NotProcessed": [], "Successful": {}, "WithMetadata": {}}
        processed, dfiles, metadata = self._collectFileDescendents(lfns, depth, production, checkreplica, False)
        return {
            "Failed": [],  # always emtpy (or error) in this implementation
            "NotProcessed": [lfn for lfn in lfns if not processed[lfn]],  # preserve original order
            "Successful": {lfn: list(dfiles[lfn]) for lfn in dfiles if processed[lfn]},
            "WithMetadata": {lfn: {dlfn: metadata[dlfn] for dlfn in dfiles[lfn]} for lfn in dfiles if processed[lfn]},
        }

    @convertToReturnValue
    def getFileDescendentsTree(
        self, lfns: list[str], depth: int = 0, production: int = 0, checkreplica: bool = True
    ) -> dict:
        """collects descendents for specified ancestors

        :param list lfns: a list of LFNs (ancestors)
        :param int depth: the depth of the processing pass chain(how far to go)
        :param productin: production number, when not zero, search descendats in that production only
        :param bool checkreplica: when set, returned descendents should have a replica
        :returns: the tree of descendents and metadata (metadata are filteres)
        """
        if not lfns:
            return {"descendents": {}, "metadata": {}}
        _, dfiles, metadata = self._collectFileDescendents(lfns, depth, production, checkreplica, True)
        if checkreplica or production != 0:
            # cleanup by removing branches to not interesting files
            # not interesting (not in metadata) leafs (not in dfiles)
            for lfn in dfiles:
                to_clean = [dlfn for dlfn in dfiles[lfn] if dlfn not in metadata and dlfn not in dfiles]
                for dlfn in to_clean:
                    del dfiles[lfn][dlfn]
            # iteratively remove not interesting branches
            while True:
                to_clean = [lfn for lfn in dfiles if lfn not in metadata and lfn not in lfns and not dfiles[lfn]]
                if not to_clean:
                    break
                for lfn in to_clean:
                    for dlfn in dfiles:
                        if lfn in dfiles[dlfn]:
                            del dfiles[dlfn][lfn]
                    del dfiles[lfn]
        for lfn in dfiles:  # make the tree from direct descendents
            dfiles[lfn].update({dlfn: dfiles[dlfn] if dlfn in dfiles else {} for dlfn in dfiles[lfn]})
        return {"descendents": {lfn: dfiles[lfn] for lfn in lfns}, "metadata": metadata}

    @convertToReturnValue
    def getFiles(
        self,
        simdesc,
        datataking,
        procPass,
        ftype,
        evt,
        configName="ALL",
        configVersion="ALL",
        production="ALL",
        flag="ALL",
        startDate=None,
        endDate=None,
        nbofEvents=False,
        startRunID=None,
        endRunID=None,
        runnumbers=None,
        replicaFlag="ALL",
        visible="ALL",
        filesize=False,
        tcks=None,
        jobStart=None,
        jobEnd=None,
        smog2States=None,
        dqok=None,
        seed_md5=None,
        sample_max=None,
    ):
        if nbofEvents:
            selection = "SUM(f.eventstat)"
        elif filesize:
            selection = "SUM(f.filesize)"
        else:
            # TODO: Ideally DISTINCT shouldn't be needed here but we should
            # probably add a unique constraint to the DB before removing it.
            # Actually, the previous two queries don't consider the DISTINCT
            # so if it's needed here those queries are incorrect.
            selection = "DISTINCT f.filename"

        return buildQuery(
            self.dbR_,
            selection,
            combineDescription(simdesc, datataking),
            procPass,
            ftype,
            evt,
            configName,
            configVersion,
            production,
            flag,
            startDate,
            endDate,
            startRunID,
            endRunID,
            runnumbers,
            replicaFlag,
            visible,
            tcks,
            jobStart,
            jobEnd,
            smog2States,
            dqok,
            seed_md5,
            sample_max,
        )

    @convertToReturnValue
    def getFilesWithMetadata(
        self,
        configName,
        configVersion,
        conddescription="ALL",
        processing="ALL",
        evt="ALL",
        production="ALL",
        filetype="ALL",
        quality="ALL",
        visible="ALL",
        replicaflag="ALL",
        startDate=None,
        endDate=None,
        runnumbers=None,
        startRunID=None,
        endRunID=None,
        tcks="ALL",
        jobStart=None,
        jobEnd=None,
        selection=None,
        smog2States=None,
        dqok=None,
        seed_md5=None,
        sample_max=None,
        *,
        parameters=None,
    ):
        if selection is not None:
            raise NotImplementedError("Selection is not implemented")

        forceJoins = []
        if parameters is None:
            parameters = list(GETFILESWITHMETADATA_NAME_TO_COL)
        selection = []
        for parameter in parameters:
            col = GETFILESWITHMETADATA_NAME_TO_COL[parameter]
            join = GETFILESWITHMETADATA_AVAILABLE_COLUMNS[col]
            selection.append(f"{TABLE_JOIN_TO_NAME[join]}.{col}")
            if join is not None:
                forceJoins.append(join)

        return buildQuery(
            self.dbR_,
            ", ".join(selection),
            conddescription,
            processing,
            filetype,
            evt,
            configName,
            configVersion,
            production,
            quality,
            startDate,
            endDate,
            startRunID,
            endRunID,
            runnumbers,
            replicaflag,
            visible,
            tcks,
            jobStart,
            jobEnd,
            smog2States,
            dqok,
            seed_md5,
            sample_max,
            forceJoins=forceJoins,
            # FIXME: Arguably this is a bug but it's needed to reproduce the behaviour of the original code
            # After we've migrated to the new method we should probably remove this?
            extraConditions={"f.eventtypeid IS NOT NULL"},
        )

    @convertToReturnValue
    def getVisibleFilesWithMetadata(
        self,
        simdesc,
        datataking,
        procPass,
        ftype,
        evt,
        configName="ALL",
        configVersion="ALL",
        production="ALL",
        flag="ALL",
        startDate=None,
        endDate=None,
        nbofEvents=None,
        startRunID=None,
        endRunID=None,
        runnumbers=None,
        replicaFlag="Yes",
        tcks=None,
        jobStart=None,
        jobEnd=None,
        smog2States=None,
        dqok=None,
        seed_md5=None,
        sample_max=None,
    ):
        # These options are broken the in the original code
        if nbofEvents is not None:
            raise NotImplementedError("nbofEvents is not implemented for visible files")
        if replicaFlag != "Yes":
            raise NotImplementedError("This option is broken in the original code")

        selection = (
            "DISTINCT f.filename, f.eventstat, j.eventinputstat, j.runnumber, j.fillnumber, f.filesize, "
            "j.totalluminosity, f.luminosity, f.instLuminosity, j.tck"
        )
        return buildQuery(
            self.dbR_,
            selection,
            combineDescription(simdesc, datataking),
            procPass,
            ftype,
            evt,
            configName,
            configVersion,
            production,
            flag,
            startDate,
            endDate,
            startRunID,
            endRunID,
            runnumbers,
            replicaFlag,
            "Y",
            tcks,
            jobStart,
            jobEnd,
            smog2States,
            dqok,
            seed_md5,
            sample_max,
            forceJoins={TableJoins.JOBS},
        )

    @convertToReturnValue
    def getFilesSummary(
        self,
        configName,
        configVersion,
        conditionDescription="ALL",
        processingPass="ALL",
        eventType="ALL",
        production="ALL",
        fileType="ALL",
        dataQuality="ALL",
        startRun="ALL",
        endRun="ALL",
        visible="ALL",
        startDate=None,
        endDate=None,
        runNumbers=None,
        replicaFlag="ALL",
        tcks="ALL",
        jobStart=None,
        jobEnd=None,
        smog2States=None,
        dqok=None,
        seed_md5=None,
        sample_max=None,
    ):
        selection = "COUNT(fileid), SUM(f.EventStat), SUM(f.FILESIZE), SUM(f.luminosity), SUM(f.instLuminosity)"
        return buildQuery(
            self.dbR_,
            selection,
            conditionDescription,
            processingPass,
            fileType,
            eventType,
            configName,
            configVersion,
            production,
            dataQuality,
            startDate,
            endDate,
            startRun,
            endRun,
            runNumbers,
            replicaFlag,
            visible,
            tcks,
            jobStart,
            jobEnd,
            smog2States,
            dqok,
            seed_md5,
            sample_max,
        )



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""DISET request handler for the LHCbDIRAC/TransformationDB."""

from DIRAC import S_OK, S_ERROR
from DIRAC.TransformationSystem.Service.TransformationManagerHandler import TransformationManagerHandler as TManagerBase


class TransformationManagerHandlerMixin:
    types_deleteTransformation = [int]

    def export_deleteTransformation(self, transID):
        rc = self.getRemoteCredentials()
        author = rc.get("DN", rc.get("CN"))
        return self.transformationDB.deleteTransformation(transID, author=author)

    types_setHotFlag = [int, bool]

    def export_setHotFlag(self, transID, hotFlag):
        return self.transformationDB.setHotFlag(transID, hotFlag)

    #############################################################################
    #
    # Managing the BkQueries table
    #

    types_addBookkeepingQuery = [int, dict]

    @classmethod
    def export_addBookkeepingQuery(self, transID, queryDict):
        return self.transformationDB.addBookkeepingQuery(transID, queryDict)

    types_deleteBookkeepingQuery = [int]

    @classmethod
    def export_deleteBookkeepingQuery(self, transID):
        return self.transformationDB.deleteBookkeepingQuery(transID)

    types_getBookkeepingQuery = [int]

    @classmethod
    def export_getBookkeepingQuery(self, transID):
        return self.transformationDB.getBookkeepingQuery(transID)

    types_getBookkeepingQueries = [list]

    @classmethod
    def export_getBookkeepingQueries(self, transIDs):
        return self.transformationDB.getBookkeepingQueries(transIDs)

    types_getTransformationsWithBkQueries = [list]

    @classmethod
    def export_getTransformationsWithBkQueries(self, transIDs):
        return self.transformationDB.getTransformationsWithBkQueries(transIDs)

    types_setBookkeepingQueryEndRun = [int, int]

    @classmethod
    def export_setBookkeepingQueryEndRun(self, transID, runNumber):
        return self.transformationDB.setBookkeepingQueryEndRun(transID, runNumber)

    types_setBookkeepingQueryStartRun = [int, int]

    @classmethod
    def export_setBookkeepingQueryStartRun(self, transID, runNumber):
        return self.transformationDB.setBookkeepingQueryStartRun(transID, runNumber)

    types_addBookkeepingQueryRunList = [int, [list]]

    @classmethod
    def export_addBookkeepingQueryRunList(self, transID, runList):
        return self.transformationDB.addBookkeepingQueryRunList(transID, runList)

    #############################################################################
    #
    # Managing the TransformationRuns table
    #

    types_getTransformationRuns = []

    @classmethod
    def export_getTransformationRuns(self, condDict={}, orderAttribute=None, limit=None):
        return self.transformationDB.getTransformationRuns(condDict, orderAttribute=orderAttribute, limit=limit)

    types_insertTransformationRun = [int, int, str]

    @classmethod
    def export_insertTransformationRun(self, transID, runID, selectedSite=""):
        return self.transformationDB.insertTransformationRun(transID, runID, selectedSite="")

    types_getTransformationRunStats = [[int, list]]

    @classmethod
    def export_getTransformationRunStats(self, transIDs):
        if isinstance(transIDs, int):
            transIDs = [transIDs]
        return self.transformationDB.getTransformationRunStats(transIDs)

    types_addTransformationRunFiles = [int, int, list]

    @classmethod
    def export_addTransformationRunFiles(self, transID, runID, lfns):
        return self.transformationDB.addTransformationRunFiles(transID, runID, lfns)

    types_setParameterToTransformationFiles = [int, dict]

    @classmethod
    def export_setParameterToTransformationFiles(self, transID, lfnsDict):
        return self.transformationDB.setParameterToTransformationFiles(transID, lfnsDict)

    types_setTransformationRunStatus = [int, [int, list], str]

    @classmethod
    def export_setTransformationRunStatus(self, transID, runID, status):
        return self.transformationDB.setTransformationRunStatus(transID, runID, status)

    types_setTransformationRunsSite = [int, int, str]

    @classmethod
    def export_setTransformationRunsSite(self, transID, runID, assignedSE):
        return self.transformationDB.setTransformationRunsSite(transID, runID, assignedSE)

    types_getTransformationRunsSummaryWeb = [dict, list, int, int]

    def export_getTransformationRunsSummaryWeb(self, selectDict, sortList, startItem, maxItems):
        """Get the summary of the transformation run information for a given page
        in the generic format."""

        # Obtain the timing information from the selectDict
        last_update = selectDict.get("LastUpdate", None)
        if last_update:
            del selectDict["LastUpdate"]
        fromDate = selectDict.get("FromDate", None)
        if fromDate:
            del selectDict["FromDate"]
        if not fromDate:
            fromDate = last_update
        toDate = selectDict.get("ToDate", None)
        if toDate:
            del selectDict["ToDate"]
        # Sorting instructions. Only one for the moment.
        if sortList:
            orderAttribute = sortList[0][0] + ":" + sortList[0][1]
        else:
            orderAttribute = None

        # Get the transformations that match the selection
        res = self.transformationDB.getTransformationRuns(
            condDict=selectDict, older=toDate, newer=fromDate, orderAttribute=orderAttribute
        )
        if not res["OK"]:
            self.log.error("TransformationManager.getTransformationRuns()", res["Message"])
            return res

        # Prepare the standard structure now within the resultDict dictionary
        resultDict = {}
        trList = res["Records"]
        # Create the total records entry
        nTrans = len(trList)
        resultDict["TotalRecords"] = nTrans
        # Create the ParameterNames entry
        paramNames = res["ParameterNames"]
        resultDict["ParameterNames"] = list(paramNames)

        # Add the job states to the ParameterNames entry
        # taskStateNames   = ['Created','Running','Submitted','Failed','Waiting','Done','Stalled']
        # resultDict['ParameterNames'] += ['Jobs_'+x for x in taskStateNames]
        # Add the file states to the ParameterNames entry
        fileStateNames = [
            "PercentProcessed",
            "Processed",
            "Unused",
            "Assigned",
            "Total",
            "Problematic",
            "ApplicationCrash",
            "MaxReset",
        ]
        resultDict["ParameterNames"] += ["Files_" + x for x in fileStateNames]

        # Get the transformations which are within the selected window
        if nTrans == 0:
            return S_OK(resultDict)
        ini = startItem
        last = ini + maxItems
        if ini >= nTrans:
            return S_ERROR("Item number out of range")
        if last > nTrans:
            last = nTrans
        transList = trList[ini:last]
        if not transList:
            return S_OK(resultDict)

        # Obtain the run statistics for the requested transformations
        transIDs = []
        for transRun in transList:
            transRunDict = dict(zip(paramNames, transRun))
            transID = int(transRunDict["TransformationID"])
            if transID not in transIDs:
                transIDs.append(transID)
        res = self.transformationDB.getTransformationRunStats(transIDs)
        if not res["OK"]:
            return res
        transRunStatusDict = res["Value"]

        statusDict = {}
        # Add specific information for each selected transformation/run
        for transRun in transList:
            transRunDict = dict(zip(paramNames, transRun))
            transID = transRunDict["TransformationID"]
            runID = transRunDict["RunNumber"]
            if transID not in transRunStatusDict or runID not in transRunStatusDict[transID]:
                for state in fileStateNames:
                    transRun.append(0)
                continue
            # Update the status counters
            status = transRunDict["Status"]
            statusDict[status] = statusDict.setdefault(status, 0) + 1

            # Populate the run file statistics
            fileDict = transRunStatusDict[transID][runID]
            if fileDict["Total"] == 0:
                fileDict["PercentProcessed"] = 0
            else:
                processed = fileDict.get("Processed", 0)
                fileDict["PercentProcessed"] = f"{int(processed * 1000.0 / fileDict['Total']) / 10.0:.1f}"
            for state in fileStateNames:
                if fileDict and state in fileDict:
                    transRun.append(fileDict[state])
                else:
                    transRun.append(0)

            # Get the statistics on the number of jobs for the transformation
            # res = database.getTransformationTaskRunStats(transID)
            # taskDict = {}
            # if res['OK'] and res['Value']:
            #  taskDict = res['Value']
            # for state in taskStateNames:
            #  if taskDict and taskDict.has_key(state):
            #    trans.append(taskDict[state])
            #  else:
            #    trans.append(0)

        resultDict["Records"] = transList
        resultDict["Extras"] = statusDict
        return S_OK(resultDict)

    #############################################################################
    #
    # Managing the RunsMetadata table
    #

    types_addRunsMetadata = [int, dict]

    @classmethod
    def export_addRunsMetadata(self, runID, metadataDict):
        """insert run metadata."""
        return self.transformationDB.setRunsMetadata(runID, metadataDict)

    types_updateRunsMetadata = [int, dict]

    @classmethod
    def export_updateRunsMetadata(self, runID, metadataDict):
        """insert run metadata."""
        return self.transformationDB.updateRunsMetadata(runID, metadataDict)

    types_getRunsMetadata = [[list, int]]

    @classmethod
    def export_getRunsMetadata(self, runID):
        """retrieve run metadata."""
        return self.transformationDB.getRunsMetadata(runID)

    types_deleteRunsMetadata = [int]

    @classmethod
    def export_deleteRunsMetadata(self, runID):
        """delete run metadata."""
        return self.transformationDB.deleteRunsMetadata(runID)

    types_getRunsInCache = [dict]

    @classmethod
    def export_getRunsInCache(self, condDict):
        """gets what's in."""
        return self.transformationDB.getRunsInCache(condDict)

    #############################################################################
    #
    # Managing the RunDestination table
    #

    types_getDestinationForRun = [[int, str, list]]

    @classmethod
    def export_getDestinationForRun(self, runIDs):
        """retrieve run destination for a single run or a list of runs."""
        if isinstance(runIDs, int):
            runIDs = [runIDs]
        if isinstance(runIDs, str):
            runIDs = [int(runIDs)]
        # expecting a list of long integers
        return self.transformationDB.getDestinationForRun(runIDs)

    types_setDestinationForRun = [int, str]

    @classmethod
    def export_setDestinationForRun(self, runID, destination):
        """set run destination."""
        return self.transformationDB.setDestinationForRun(runID, destination)

    #############################################################################
    #
    # Managing the StoredJobDescription table
    #

    types_addStoredJobDescription = [int, str]

    @classmethod
    def export_addStoredJobDescription(self, transformationID, jobDescription):
        return self.transformationDB.addStoredJobDescription(transformationID, jobDescription)

    types_getStoredJobDescription = [int]

    @classmethod
    def export_getStoredJobDescription(self, transformationID):
        return self.transformationDB.getStoredJobDescription(transformationID)

    types_removeStoredJobDescription = [int]

    @classmethod
    def export_removeStoredJobDescription(self, transformationID):
        return self.transformationDB.removeStoredJobDescription(transformationID)

    types_getStoredJobDescriptionIDs = []

    @classmethod
    def export_getStoredJobDescriptionIDs(self):
        return self.transformationDB.getStoredJobDescriptionIDs()


class TransformationManagerHandler(TransformationManagerHandlerMixin, TManagerBase):
    pass



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""UserStorage Type."""
from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType


class UserStorage(BaseAccountingType):
    """UserStorage as extension of BaseAccountingType."""

    def __init__(self):
        super().__init__()

        self.definitionKeyFields = [("StorageElement", "VARCHAR(32)"), ("User", "VARCHAR(32)")]

        self.definitionAccountingFields = [
            ("LogicalSize", "BIGINT UNSIGNED"),
            ("LogicalFiles", "BIGINT UNSIGNED"),
            ("PhysicalSize", "BIGINT UNSIGNED"),
            ("PhysicalFiles", "BIGINT UNSIGNED"),
            ("StorageSize", "BIGINT UNSIGNED"),
            ("StorageFiles", "BIGINT UNSIGNED"),
        ]
        self.bucketsLength = [
            (86400 * 30 * 6, 86400),  # <6m = 1d
            (31104000, 604800),  # >6m = 1w
        ]
        self.checkType()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the visibility flag to a dataset."""

from DIRAC import exit as dExit, gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerFileSwitches()

    Script.registerSwitch("", "List", "   Print out the list of LFNs")
    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)
    dumpList = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "List":
            dumpList = True

    bkQuery = dmScript.getBKQuery()
    lfns = dmScript.getOption("LFNs", [])
    if not bkQuery and not lfns:
        gLogger.error("No BKQuery and no files given...")
        dExit(1)
    # Invert the visibility flag as want to set Invisible those that are visible and vice-versa
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bk = BookkeepingClient()

    visibilityFlag = dmScript.getOption("Visibility", None)
    if visibilityFlag is None:
        gLogger.error("Visibility option should be given")
        dExit(2)
    visibilityFlag = str(visibilityFlag).lower() == "yes"
    if bkQuery:
        # Query with visibility opposite to what is requested to be set ;-)
        bkQuery.setOption("Visible", "No" if visibilityFlag else "Yes")
        gLogger.notice("BQ query:", bkQuery)
        lfns += bkQuery.getLFNs()
    if not lfns:
        gLogger.notice("No files found...")
    else:
        res = {"OK": True}
        if visibilityFlag:
            res = bk.setFilesVisible(lfns)
            msg = "visible"
        else:
            res = bk.setFilesInvisible(lfns)
            msg = "invisible"
        if not res["OK"]:
            gLogger.error("Error setting files", msg)
            dExit(1)
        gLogger.notice(f"Successfully set {len(lfns)} files {msg + (':' if dumpList else '')}")
        if dumpList:
            gLogger.notice("\n".join(lfns))


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""stores the replica information."""
from DIRAC import gLogger


class ReplicaParam:
    """ReplicaParam class."""

    def __init__(self):
        """initialize the member of the class."""
        self.file = ""
        self.name = ""
        self.location = ""
        self.se = ""
        self.action = ""

    def __repr__(self):
        """formats the output of print."""
        result = "\n Replica:\n"
        result += self.file + " " + self.name + " " + self.location + " "
        result += self.se + " " + self.action

        return result

    def writeToXML(self, flag=True):
        """creates an XML string."""
        # job replica param
        gLogger.info("replica param", str(flag))
        if flag:
            result = '     <Replica Name="' + self.name + '" Location="' + self.location + '"/>\n'

        else:
            result = '<Replica File="' + self.file + '"\n'
            result += '      Name="' + self.name + '"\n'
            result += '      Location="' + self.location + '"\n'
            result += '      SE="' + self.se + '"/> \n'

        return result



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Does a TS -> BK check for processed files with descendants."""
import sys
import os
import time

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()
    depth = 1

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [ProdIDs]",
            ]
        )
    )
    Script.registerSwitch("", "Runs=", "   Specify the run range")
    Script.registerSwitch(
        "", "ActiveRunsProduction=", "   Specify the production from which the runs should be derived"
    )
    Script.registerSwitch("", "FileType=", "S   pecify the descendants file type")
    Script.registerSwitch("", "NoFC", "   Trust the BK replica flag, no FC check")
    Script.registerSwitch("", "FixIt", "   Fix the files in transformation table")
    Script.registerSwitch("", "Verbose", "   Print full list of files with error")
    Script.registerSwitch("", "Status=", "   Select files with a given status in the production")
    Script.registerSwitch("", "Depth=", "   Depth to which to check descendants (default=%d)" % depth)
    Script.registerSwitch("", "Force", "   Use this flag to force checking a whole production")
    Script.parseCommandLine(ignoreErrors=True)
    fileType = []
    runsList = []
    lfnList = []
    fixIt = False
    fromProd = None
    verbose = False
    status = None
    noFC = False
    force = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Runs":
            try:
                runs = switch[1].split(",")
                runsList = []
                for run in runs:
                    runRange = run.split(":")
                    if len(runRange) == 2:
                        runsList += list(range(int(runRange[0]), int(runRange[1]) + 1))
                    else:
                        runsList.append(int(run))
            except Exception as e:
                gLogger.exception("Bad run range", switch[1], lException=e)
                DIRAC.exit(1)
        elif switch[0] == "Status":
            status = switch[1].split(",")
        elif switch[0] == "Verbose":
            verbose = True
        elif switch[0] == "FileType":
            fileType = switch[1].split(",")
        elif switch[0] == "FixIt":
            fixIt = True
        elif switch[0] == "NoFC":
            noFC = True
        elif switch[0] == "Depth":
            depth = min(10, max(1, int(switch[1])))
        elif switch[0] == "Force":
            force = True
        elif switch[0] == "ActiveRunsProduction":
            try:
                fromProd = int(switch[1])
            except ValueError as e:
                gLogger.exception(f"Wrong production number: {switch[1]}", lException=e)
                DIRAC.exit(0)

    args = Script.getPositionalArgs()
    if not len(args):
        gLogger.error("Specify transformation number...")
        DIRAC.exit(0)
    else:
        ids = args[0].split(",")
        prodList = []
        try:
            for id in ids:
                r = id.split(":")
                if len(r) > 1:
                    for i in range(int(r[0]), int(r[1]) + 1):
                        prodList.append(i)
                else:
                    prodList.append(int(eval(r[0])))
        except (ValueError, NameError) as e:
            gLogger.exception(f"Bad production list: {args[0]}", lException=e)
    # In case the user asked for specific LFNs
    if not status:
        lfnList = dmScript.getOption("LFNs", [])

    if not status and not lfnList and not runsList and not fromProd and not force:
        gLogger.fatal("You are about to check descendants for all files in a production")
        gLogger.fatal("If you really want to do so, use --Force")
        DIRAC.exit(0)

    from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
    from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    tr = TransformationClient()
    for prod in prodList:
        startTime = time.time()
        cc = ConsistencyChecks()
        # Setting the prod also sets its type
        try:
            cc.prod = prod
        except RuntimeError as e:
            gLogger.exception(lException=e)
            continue
        if fileType and cc.transType in ("Merge", "MCMerge"):
            gLogger.notice("It is not allowed to select file type for merging transformation", prod)
            continue
        cc.verbose = verbose
        cc.noFC = noFC
        cc.descendantsDepth = depth
        if prod != prodList[0]:
            gLogger.notice("====================")
        gLogger.notice("Processing %s production %d" % (cc.transType, cc.prod))

        if status:
            cc.status = status
        if lfnList:
            cc.lfns = lfnList
        if not fileType:
            bkQuery = BKQuery({"Production": prod, "FileType": "ALL", "Visible": "All"})
            cc.fileType = bkQuery.getBKFileTypes()
            gLogger.notice(f"Looking for descendants of type {str(cc.fileType)}")
            notAllFileTypes = False
        else:
            cc.fileType = fileType
            cc.fileTypesExcluded = ["LOG"]
            notAllFileTypes = True
        cc.runsList = runsList
        cc.runStatus = "Active"
        cc.fromProd = fromProd
        cc.checkTS2BK()

        # Print out the results
        gLogger.notice("\nResults:")
        if verbose:
            nMax = sys.maxsize
        else:
            nMax = 20
        suffix = ""
        n = 0
        while True:
            fileName = f"CheckDescendantsResults_{str(cc.prod)}{suffix}.txt"
            if not os.path.exists(fileName):
                break
            n += 1
            suffix = "-%d" % n
        fp = None
        if cc.inFCNotInBK:
            lfns = cc.inFCNotInBK
            gLogger.notice(f"{len(lfns)} descendants were found in FC but don't have replica flag in BK")
            if not fp:
                fp = open(fileName, "w")
            fp.write("\nInFCNotInBK ".join([""] + lfns))
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            gLogger.notice(
                "To fix it:   grep InFCNotInBK %s | dirac-dms-check-fc2bkk\n"
                "\tBeware you can then either fix the BK or the FC/SE" % fileName
            )

        if cc.inFailover:
            lfns = cc.inFailover
            gLogger.notice(f"{len(lfns)} descendants were found in Failover and have no replica flag in BK")
            if not fp:
                fp = open(fileName, "w")
            fp.write("\nInFailover ".join([""] + lfns))
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            gLogger.notice(
                "You should check whether they are in a failover request by looking "
                "at their job status and in the RMS..."
            )
            gLogger.notice(f"To list them:     grep InFailover {fileName}")

        if cc.inBKNotInFC:
            lfns = cc.inBKNotInFC
            gLogger.notice(f"{len(lfns)} descendants were found with replica flag in BK but not in FC")
            if not fp:
                fp = open(fileName, "w")
            fp.write("\nInBKNotInFC ".join([""] + lfns))
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            gLogger.notice(f"To try and fix this:    grep InBKNotInFC {fileName} | dirac-dms-check-bkk2fc")

        if cc.removedFiles:
            from DIRAC.Core.Utilities.List import breakListIntoChunks

            gLogger.notice(
                "%d input files are Processed, have no descendants but are no longer in the FC\n"
                "  As they cannot be reset Unused, set them Removed" % len(cc.removedFiles)
            )
            if not fp:
                fp = open(fileName, "w")
            fp.write("\nProcNotinFC ".join([""] + cc.removedFiles))
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(cc.removedFiles) > nMax else "All files:",
                "\n".join([""] + cc.removedFiles[0:nMax]),
            )
            for lfnChunk in breakListIntoChunks(cc.removedFiles, 1000):
                while True:
                    res = cc.transClient.setFileStatusForTransformation(cc.prod, "Removed", lfnChunk, force=True)
                    if not res["OK"]:
                        gLogger.notice("Error setting files Removed, retry...", res["Message"])
                    else:
                        break
            gLogger.notice("\tFiles successfully set to status Removed")

        gLogger.notice(
            "%d unique daughters found with real descendants"
            % (len(set(cc.descForPrcdLFNs).union(cc.descForNonPrcdLFNs)))
        )

        if cc.prcdWithMultDesc:
            lfns = sorted(cc.prcdWithMultDesc)
            gLogger.notice(f"Processed LFNs with multiple descendants ({len(lfns)}) -> ERROR")
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            if not fp:
                fp = open(fileName, "w")
            fp.write(
                "\nProcMultDesc ".join([""] + [f"{lfn}: {str(multi)}" for lfn, multi in cc.prcdWithMultDesc.items()])
            )
            gLogger.notice("I'm not doing anything for them, neither with the 'FixIt' option")
        else:
            gLogger.notice("No processed LFNs with multiple descendants found -> OK!")

        fixItUsed = False
        if cc.prcdWithoutDesc:
            lfns = set(cc.prcdWithoutDesc)
            res = cc.bkClient.getFileMetadata(list(lfns))
            badLfns = set()
            if res["OK"]:
                # Check the DQ flag of these files
                badLfns = {lfn for lfn, meta in res["Value"]["Successful"].items() if meta["DataqualityFlag"] == "BAD"}
            if badLfns:
                gLogger.notice(f"Processed LFNs without descendants ({len(lfns)})")
                gLogger.notice(
                    "WARNING: %d of these files have a BAD data quality flag and thus may have been removed"
                    % len(badLfns)
                )
                lfns -= badLfns
            if fixIt and lfns:
                fixIt = False
                gLogger.notice(f"Processed LFNs without descendants ({len(lfns)}) -> ERROR!")
                gLogger.notice("Resetting them 'Unused'")
                res = cc.transClient.setFileStatusForTransformation(prod, "Unused", list(lfns), force=True)
                if not res["OK"]:
                    gLogger.notice("Error resetting files to Unused", res["Message"])
            else:
                fixItUsed = True
                if not fp:
                    fp = open(fileName, "w")
                if badLfns:
                    fp.write("\nProcButBAD ".join([""] + sorted(badLfns)))
                    gLogger.notice("BAD files processed without descendants can be checked using:")
                    gLogger.notice(f"     grep ProcButBAD {fileName}")
                if lfns:
                    lfns = sorted(lfns)
                    gLogger.notice(f"Processed LFNs without descendants ({len(lfns)}) -> ERROR!")
                    gLogger.notice(
                        "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                        "\n".join([""] + lfns[0:nMax]),
                    )
                    fp.write("\nProcNoDesc ".join([""] + lfns))
                    if notAllFileTypes:
                        gLogger.notice("You may want to check those files again for all file types, using:")
                        gLogger.notice(
                            "     grep ProcNoDesc %s | dirac-production-check-descendants %d" % (fileName, cc.prod)
                        )
                    if cc.transType in ("DataStripping", "MCReconstruction", "MCStripping"):
                        nextProd, nextType = cc._findNextProduction()
                        if nextProd:
                            gLogger.notice(
                                "This is a %s production, thus "
                                "the problem may come from the %s production %d" % (cc.transType, nextType, nextProd)
                            )
                            gLogger.notice("     You may check it with this command:")
                            gLogger.notice(
                                "     grep ProcNoDesc %s | dirac-bookkeeping-get-file-descendants --All | grep %d |"
                                "dirac-production-check-descendants %d " % (fileName, cc.prod, nextProd)
                            )
                    gLogger.notice("If you are sure, use --FixIt for resetting those files Unused in TS")
                else:
                    gLogger.notice("No files processed without descendants that are not BAD")
        else:
            gLogger.notice("No processed LFNs without descendants found -> OK!")

        if cc.nonPrcdWithMultDesc:
            lfns = sorted(cc.nonPrcdWithMultDesc)
            gLogger.notice(f"Non processed LFNs with multiple descendants ({len(lfns)}) -> ERROR")
            if not fp:
                fp = open(fileName, "w")
            fp.write("\nNotProcMultDesc ".join([""] + lfns))
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            gLogger.notice("I'm not doing anything for them, neither with the 'FixIt' option")
        else:
            gLogger.notice("No non processed LFNs with multiple descendants found -> OK!")

        # fixing, if requested
        if cc.nonPrcdWithDesc:
            lfns = sorted(cc.nonPrcdWithDesc)
            gLogger.notice(f"There are {len(lfns)} LFNs not marked Processed but that have descendants -> ERROR")
            gLogger.notice(
                "First %d files:" % nMax if not verbose and len(lfns) > nMax else "All files:",
                "\n".join([""] + lfns[0:nMax]),
            )
            if fixIt:
                fixIt = False
                gLogger.notice("Marking them as 'Processed'")
                cc.transClient.setFileStatusForTransformation(prod, "Processed", lfns, force=True)
            else:
                if not fp:
                    fp = open(fileName, "w")
                fp.write("\nNotProcWithDesc ".join([""] + lfns))
                if not fixItUsed:
                    gLogger.notice("Use --FixIt for setting files Processed in TS")
                else:
                    gLogger.notice("To get the list of files and check them again:")
                    gLogger.notice(f"     grep NotProcWithDesc {fileName} | dirac-production-check-descendants")
        else:
            gLogger.notice("No non processed LFNs with descendants found -> OK!")
        if fp:
            fp.close()
            gLogger.notice(f"Complete list of files is in {fileName}")
        gLogger.notice("Processed production %d in %.1f seconds" % (cc.prod, time.time() - startTime))


if __name__ == "__main__":
    main()
